import logging
import sys
import time
from queue import Queue
from threading import Thread
from typing import TYPE_CHECKING, Callable, List, Optional, Type, Union

import angr
from angr.analyses.disassembly import Instruction
from angr.block import Block
from angr.knowledge_base import KnowledgeBase
from angr.knowledge_plugins import Function
from angr.misc.testing import is_testing
from cle import SymbolType

from angrmanagement.data.breakpoint import Breakpoint, BreakpointManager, BreakpointType
from angrmanagement.data.trace import Trace
from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.debugger import DebuggerListManager, DebuggerManager
from angrmanagement.logic.debugger.simgr import SimulationDebugger
from angrmanagement.logic.threads import gui_thread_schedule, gui_thread_schedule_async
from angrmanagement.ui.dialogs import AnalysisOptionsDialog

from .analysis_options import (
    AnalysesConfiguration,
    CFGAnalysisConfiguration,
    FlirtAnalysisConfiguration,
    VariableRecoveryConfiguration,
)
from .jobs import (
    CFGGenerationJob,
    CodeTaggingJob,
    FlirtSignatureRecognitionJob,
    PrototypeFindingJob,
    VariableRecoveryJob,
)
from .log import LogRecord, initialize
from .object_container import ObjectContainer

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace

_l = logging.getLogger(__name__)


class Instance:
    """
    An object to give access to normal angr project objects like a Project, CFG, and other analyses.
    """

    project: ObjectContainer
    cfg: Union[angr.analyses.cfg.CFGBase, ObjectContainer]
    cfb: Union[angr.analyses.cfg.CFBlanket, ObjectContainer]
    log: Union[List[LogRecord], ObjectContainer]

    def __init__(self):
        # pylint:disable=import-outside-toplevel
        # delayed import
        from angrmanagement.ui.views.interaction_view import (
            BackslashTextProtocol,
            PlainTextProtocol,
            ProtocolInteractor,
            SavedInteraction,
        )

        self._live = False
        self.workspace: Optional["Workspace"] = None
        self.variable_recovery_job: Optional[VariableRecoveryJob] = None
        self._analysis_configuration = None

        self.jobs = []
        self._jobs_queue = Queue()
        self.current_job = None
        self.worker_thread = None

        self.extra_containers = {}
        self._container_defaults = {}

        # where this binary is coming from - if it's loaded from a URL, then original_binary_path will be the URL
        self.original_binary_path = None
        # where this binary is now - if it's loaded from a URL, then binary_path will be its temporary location on the
        # local machine
        self.binary_path = None
        self.register_container("project", lambda: None, Optional[angr.Project], "The current angr project")
        self.register_container("simgrs", lambda: [], List[angr.SimulationManager], "Global simulation managers list")
        self.register_container("states", lambda: [], List[angr.SimState], "Global states list")
        self.register_container("patches", lambda: None, None, "Global patches update notifier")  # dummy
        self.register_container("cfg", lambda: None, Optional[angr.knowledge_plugins.cfg.CFGModel], "The current CFG")
        self.register_container("cfb", lambda: None, Optional[angr.analyses.cfg.CFBlanket], "The current CFBlanket")
        self.register_container("interactions", lambda: [], List[SavedInteraction], "Saved program interactions")
        # TODO: the current setup will erase all loaded protocols on a new project load! do we want that?
        self.register_container(
            "interaction_protocols",
            lambda: [PlainTextProtocol, BackslashTextProtocol],
            List[Type[ProtocolInteractor]],
            "Available interaction protocols",
        )
        self.register_container("log", lambda: [], List[LogRecord], "Saved log messages", logging_permitted=False)
        self.register_container("current_trace", lambda: None, Type[Trace], "Currently selected trace")
        self.register_container("traces", lambda: [], List[Trace], "Global traces list")

        self.register_container("active_view_state", lambda: None, "ViewState", "Currently focused view state")

        self.breakpoint_mgr = BreakpointManager()
        self.debugger_list_mgr = DebuggerListManager()
        self.debugger_mgr = DebuggerManager(self.debugger_list_mgr)

        self.simgrs.am_subscribe(self._update_simgr_debuggers)
        self.project.am_subscribe(self.initialize)

        # Callbacks
        self._insn_backcolor_callback: Optional[Callable[[int, bool], None]] = None  # (addr, is_selected)
        self._label_rename_callback: Optional[Callable[[int, str], None]] = None  # (addr, new_name)
        self._set_comment_callback: Optional[Callable[[int, str], None]] = None  # (addr, comment_text)

        # Setup logging
        initialize(self)

        self.cfg_args = None
        self.variable_recovery_args = None
        self._disassembly = {}
        self.pseudocode_variable_kb = None

        self._start_worker()

        self.database_path = None

        # The image name when loading image
        self.img_name = None

        self._live = True

    #
    # Properties
    #

    @property
    def kb(self) -> Optional[angr.KnowledgeBase]:
        if self.project.am_none:
            return None
        return self.project.kb

    def __getattr__(self, k):
        if k == "extra_containers":
            return {}

        try:
            return self.extra_containers[k]
        except KeyError:
            return super().__getattribute__(k)

    def __setattr__(self, k, v):
        if k in self.extra_containers:
            self.extra_containers[k].am_obj = v
        else:
            super().__setattr__(k, v)

    def __dir__(self):
        return list(super().__dir__()) + list(self.extra_containers)

    @property
    def insn_backcolor_callback(self):
        return self._insn_backcolor_callback

    @insn_backcolor_callback.setter
    def insn_backcolor_callback(self, v):
        self._insn_backcolor_callback = v

    @property
    def label_rename_callback(self):
        return self._label_rename_callback

    @label_rename_callback.setter
    def label_rename_callback(self, v):
        self._label_rename_callback = v

    @property
    def set_comment_callback(self):
        return self._set_comment_callback

    @set_comment_callback.setter
    def set_comment_callback(self, v):
        self._set_comment_callback = v

    #
    # Public methods
    #

    def register_container(self, name, default_val_func, ty, description, **kwargs):
        if name in self.extra_containers:
            cur_ty = self._container_defaults[name][1]
            if ty != cur_ty:
                raise Exception(f"Container {name} already registered with different type: {ty} != {cur_ty}")

        else:
            self._container_defaults[name] = (default_val_func, ty)
            self.extra_containers[name] = ObjectContainer(default_val_func(), description, **kwargs)

    def initialize(self, initialized=False, **kwargs):  # pylint:disable=unused-argument
        if self.project.am_none:
            return

        if not initialized:
            if self.pseudocode_variable_kb is None:
                self.initialize_pseudocode_variable_kb()

            gui_thread_schedule_async(self.run_analysis)

        self.workspace.plugins.handle_project_initialization()

    def initialize_pseudocode_variable_kb(self):
        self.pseudocode_variable_kb = KnowledgeBase(self.project.am_obj, name="pseudocode_variable_kb")

    def add_job(self, job):
        self.jobs.append(job)
        self._jobs_queue.put(job)

    def get_instruction_text_at(self, addr):
        """
        Get the text representation of an instruction at `addr`.

        :param int addr:    Address of the instruction.
        :return:            Text representation of the instruction, or None if no instruction can be found there.
        :rtype:             Optional[str]
        """

        if self.cfb is None:
            return None

        try:
            _, obj = self.cfb.floor_item(addr)
        except KeyError:
            # no object before addr exists
            return None

        if isinstance(obj, Block):
            for insn in obj.capstone.insns:
                if insn.address == addr:
                    insn_piece = Instruction(insn, None, project=self.project)
                    return insn_piece.render()[0]
        return None

    def interrupt_current_job(self):
        """Notify the current running job that the user requested an interrupt. The job may ignore it."""
        # Due to thread scheduling, current_job reference *must* first be saved on the stack. Accessing self.current_job
        # multiple times will lead to a race condition.
        current_job = self.current_job
        if current_job:
            current_job.keyboard_interrupt()

    def join_all_jobs(self, wait_period=2.0):
        """
        Wait until self.jobs is empty for at least `wait_period` seconds.

        This is because one job may add another job upon completion. We cannot simply wait until self.jobs becomes
        empty.
        """

        last_has_job = time.time()
        while time.time() - last_has_job <= wait_period:
            while self.jobs:
                last_has_job = time.time()
                time.sleep(0.05)

    def append_code_to_console(self, hook_code_string):
        console = self.workspace._get_or_create_console_view()
        console.set_input_buffer(hook_code_string)

    def delete_hook(self, addr):
        self.project.unhook(addr)

    def add_breakpoint(self, obj: Union[str, int], type_: Optional[str] = None, size: Optional[int] = None):
        """
        Convenience function to add a breakpoint.

        Examples:
        - `workspace.add_breakpoint(0x1234)` sets an execution breakpoint on address 0x1234
        - `workspace.add_breakpoint('main')` sets an execution breakpoint on `main` function
        - `workspace.add_breakpoint('global_value')` sets a write breakpoint on `global_value`
        - `workspace.add_breakpoint('global_value', 'read', 1)` sets a 1-byte read breakpoint on `global_value`
        """
        if type(obj) is int:
            addr = obj
        elif type(obj) is str:
            sym = self.project.loader.find_symbol(obj)
            if sym is None:
                _l.error("Couldn't resolve '%s'", obj)
                return
            addr = sym.rebased_addr
            if not size:
                size = sym.size
            if not type_:
                if sym.type == SymbolType.TYPE_FUNCTION:
                    type_ = "execute"
                else:
                    type_ = "write"
        elif type(obj) is Function:
            addr = obj.addr
            if not type_:
                type_ = "execute"
        else:
            _l.error("Unexpected target object type. Expected int | str | Function")
            return

        if not size:
            size = 1

        bp_type_map = {
            None: BreakpointType.Execute,
            "execute": BreakpointType.Execute,
            "write": BreakpointType.Write,
            "read": BreakpointType.Read,
        }
        if type_ not in bp_type_map:
            _l.error("Unknown breakpoint type '%s'. Expected %s", type_, " | ".join(bp_type_map.keys()))
            return

        bp = Breakpoint(bp_type_map[type_], addr, size)
        self.breakpoint_mgr.add_breakpoint(bp)

    def set_comment(self, addr, comment_text):
        kb = self.project.kb
        exists = addr in kb.comments

        # callback
        if comment_text is None and exists:
            self.workspace.plugins.handle_comment_changed(addr, "", False, False, False)
            del kb.comments[addr]
        else:
            self.workspace.plugins.handle_comment_changed(addr, comment_text, not exists, False, False)
            kb.comments[addr] = comment_text

        # TODO: can this be removed?
        if self.set_comment_callback:
            self.set_comment_callback(addr=addr, comment_text=comment_text)

    def run_analysis(self, prompt_for_configuration=True):
        if self.project.am_none:
            return

        if self._analysis_configuration is None:
            self._analysis_configuration = AnalysesConfiguration(
                [
                    a(self)
                    for a in [CFGAnalysisConfiguration, FlirtAnalysisConfiguration, VariableRecoveryConfiguration]
                ],
                self,
            )

        if not self.workspace.main_window.shown_at_start:
            # If we are running headlessly (e.g. tests), just run with default configuration
            prompt_for_configuration = False

        if prompt_for_configuration:
            dlg = AnalysisOptionsDialog(self._analysis_configuration, self.workspace, self.workspace.main_window)
            dlg.setModal(True)
            should_run = dlg.exec_()
        else:
            should_run = True

        if should_run:
            if self._analysis_configuration["cfg"].enabled:
                self.generate_cfg(self._analysis_configuration["cfg"].to_dict())

    #
    # Private methods
    #

    @staticmethod
    def _start_daemon_thread(target, name, args=None):
        t = Thread(target=target, name=name, args=args if args else tuple())
        t.daemon = True
        t.start()
        return t

    def _start_worker(self):
        self.worker_thread = self._start_daemon_thread(self._worker, "angr-management Worker Thread")

    def _worker(self):
        while True:
            if self._jobs_queue.empty():
                gui_thread_schedule(GlobalInfo.main_window.progress_done, args=())

            if self.workspace is not None and any(job.blocking for job in self.jobs):
                gui_thread_schedule(self.workspace.main_window._progress_dialog.hide, args=())

            job = self._jobs_queue.get()
            gui_thread_schedule_async(GlobalInfo.main_window.progress, args=("Working...", 0.0))

            if any(job.blocking for job in self.jobs):
                if self.workspace.main_window.isVisible():
                    gui_thread_schedule(self.workspace.main_window._progress_dialog.show, args=())

            try:
                self.current_job = job
                result = job.run(self)
                self.current_job = None
            except (Exception, KeyboardInterrupt) as e:  # pylint: disable=broad-except
                sys.last_traceback = e.__traceback__
                self.current_job = None
                self.workspace.log('Exception while running job "%s":' % job.name)
                self.workspace.log(e)
                self.workspace.log("Type %debug to debug it")
            else:
                gui_thread_schedule_async(job.finish, args=(self, result))

    # pylint:disable=no-self-use
    def _set_status(self, status_text):
        GlobalInfo.main_window.status = status_text

    def _reset_containers(self, **kwargs):
        # pylint:disable=consider-using-dict-items
        for name in self.extra_containers:
            self.extra_containers[name].am_obj = self._container_defaults[name][0]()
            self.extra_containers[name].am_event(**kwargs)

        for dbg in list(self.debugger_list_mgr.debugger_list):
            self.debugger_list_mgr.remove_debugger(dbg)

        self.breakpoint_mgr.clear()

    def _update_simgr_debuggers(self, **kwargs):  # pylint:disable=unused-argument
        sim_dbg = None
        for dbg in self.debugger_list_mgr.debugger_list:
            if isinstance(dbg, SimulationDebugger):
                sim_dbg = dbg
                break

        if len(self.simgrs) > 0:
            if sim_dbg is None:
                view = self.workspace._get_or_create_symexec_view()._simgrs
                dbg = SimulationDebugger(view, self.workspace)
                self.debugger_list_mgr.add_debugger(dbg)
                self.debugger_mgr.set_debugger(dbg)
        elif sim_dbg is not None:
            self.debugger_list_mgr.remove_debugger(sim_dbg)

    #
    # Events
    #

    def generate_cfg(self, cfg_args=None):
        if cfg_args is None:
            cfg_args = {}

        cfg_job = CFGGenerationJob(on_finish=self.on_cfg_generated, **cfg_args)
        self.add_job(cfg_job)
        self._start_daemon_thread(self._refresh_cfg, "Progressively Refreshing CFG", args=(cfg_job,))

    def _refresh_cfg(self, cfg_job):
        """
        Reload once and then refresh in a loop, while the CFG job is running
        """
        reloaded = False
        while True:
            if not self.cfg.am_none:
                if reloaded:
                    gui_thread_schedule_async(
                        self.workspace.refresh,
                        kwargs={
                            "categories": ["disassembly", "functions"],
                        },
                    )
                else:
                    gui_thread_schedule_async(
                        self.workspace.reload,
                        kwargs={
                            "categories": ["disassembly", "functions"],
                        },
                    )
                    reloaded = True

            time.sleep(0.3)
            if cfg_job not in self.jobs:
                break

    def on_cfg_generated(self):
        if self._analysis_configuration["flirt"].enabled:
            self.add_job(
                FlirtSignatureRecognitionJob(
                    on_finish=self._on_flirt_signature_recognized,
                )
            )

        # display the main function if it exists, otherwise display the function at the entry point
        if self.cfg is not None:
            the_func = self.kb.functions.function(name="main")
            if the_func is None:
                the_func = self.kb.functions.function(addr=self.project.entry)

            if the_func is not None:
                self.on_function_selected(the_func)

            # Initialize the linear viewer
            if len(self.workspace.view_manager.views_by_category["disassembly"]) == 1:
                view = self.workspace.view_manager.first_view_in_category("disassembly")
            else:
                view = self.workspace.view_manager.current_view_in_category("disassembly")
            if view is not None:
                view._linear_viewer.initialize()

            # Reload the pseudocode view
            view = self.workspace.view_manager.first_view_in_category("pseudocode")
            if view is not None:
                view.reload()

            # Reload the strings view
            view = self.workspace.view_manager.first_view_in_category("strings")
            if view is not None:
                view.reload()

            # Clear the proximity view
            view = self.workspace.view_manager.first_view_in_category("proximity")
            if view is not None:
                view.clear()

    def _on_flirt_signature_recognized(self):
        self.add_job(
            PrototypeFindingJob(
                on_finish=self._on_prototype_found,
            )
        )

    def _on_prototype_found(self):
        self.add_job(
            CodeTaggingJob(
                on_finish=self.on_function_tagged,
            )
        )

        if self._analysis_configuration["varec"].enabled:
            options = self._analysis_configuration["varec"].to_dict()
            if is_testing:
                # disable multiprocessing on angr CI
                options["workers"] = 0
            self.variable_recovery_job = VariableRecoveryJob(
                **self._analysis_configuration["varec"].to_dict(),
                on_variable_recovered=self.on_variable_recovered,
            )
            # prioritize the current function in display
            disassembly_view = self.workspace.view_manager.first_view_in_category("disassembly")
            if disassembly_view is not None:
                if not disassembly_view.function.am_none:
                    self.variable_recovery_job.prioritize_function(disassembly_view.function.addr)
            self.add_job(self.variable_recovery_job)

    def on_function_selected(self, func: Function):
        """
        Callback function triggered when a new function is selected in the function view.

        :param func:    The function that is selected.
        :return:        None
        """

        # Ask all current views to display this function
        current_view = self.workspace.view_manager.current_tab
        if current_view is None or not current_view.FUNCTION_SPECIFIC_VIEW:
            # we don't have a current view or the current view does not have function-specific content. create a
            # disassembly view to display the selected function.
            disasm_view = self.workspace._get_or_create_disassembly_view()
            disasm_view.display_function(func)
            self.workspace.view_manager.raise_view(disasm_view)
        else:
            # ask the current view to display this function
            current_view.function = func

    def on_function_tagged(self):
        # reload disassembly view
        if len(self.workspace.view_manager.views_by_category["disassembly"]) == 1:
            view = self.workspace.view_manager.first_view_in_category("disassembly")
        else:
            view = self.workspace.view_manager.current_view_in_category("disassembly")

        if view is not None:
            if view.current_function.am_obj is not None:
                view.reload()

    def on_variable_recovered(self, func_addr: int):
        """
        Called when variable information of the given function is available.

        :param int func_addr:   Address of the function whose variable information is available.
        """
        disassembly_view = self.workspace.view_manager.first_view_in_category("disassembly")
        if disassembly_view is not None:
            disassembly_view.on_variable_recovered(func_addr)



from typing import Optional

from angrmanagement.logic.threads import gui_thread_schedule_async, is_gui_thread
from angrmanagement.plugins.base_plugin import BasePlugin

from .ui import ComponentsView, LoadComponentsDialog


class ComponentsPlugin(BasePlugin):
    REQUIRE_WORKSPACE = False

    """
    Implement a component viewer that takes JSON messages on function clustering from the server side and visualizes
    the clusters in a treeview.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.workspace is not None:
            self.view: ComponentsView = ComponentsView(self.workspace, "left")

            # register a new view
            self.workspace.view_manager.add_view(self.view)

    def teardown(self):
        pass

    #
    # Menus
    #

    MENU_BUTTONS = [
        "Load components...",
        "Reset components",
    ]
    LOAD_COMPONENTS = 0
    RESET_COMPONENTS = 1

    def handle_click_menu(self, idx):
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return

        if self.workspace.main_instance.project.am_none:
            return

        mapping = {
            self.LOAD_COMPONENTS: self.load_components,
            self.RESET_COMPONENTS: self.reset_components,
        }
        mapping.get(idx)()

    def load_components(self, url: Optional[str] = None):
        """
        Open a new dialog and take a JSON URL or a file path. Then load components from that URL.
        """
        dialog = LoadComponentsDialog(workspace=self.workspace, url=url)
        dialog.exec_()
        if dialog.tree is not None:
            self.view.load(dialog.tree)

    def reset_components(self):
        """
        Clear existing components information.
        """
        self.view.reset()

    #
    # URLs
    #

    # register actions
    URL_ACTIONS = [
        "bughouse_component",
    ]

    def handle_url_action(self, action, kwargs):
        mapping = {
            "bughouse_component": self.handle_url_action_bughouse_component,
        }

        func = mapping.get(action)
        if is_gui_thread():
            func(**kwargs)
        else:
            gui_thread_schedule_async(func, kwargs=kwargs)

    def handle_url_action_bughouse_component(self, url=None):
        self.load_components(url)



from .chess_connector import ChessConnector
from .chess_url_handler import ChessUrlHandler
from .poi_plugin import POIViewer

__all__ = [
    "ChessConnector",
    "ChessUrlHandler",
    "POIViewer",
]



import binascii
import os
from typing import Dict, Optional, Tuple

import archinfo
from cle import Blob
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFrame,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMessageBox,
    QSizePolicy,
    QTabWidget,
    QVBoxLayout,
)

try:
    import pypcode
except ImportError:
    pypcode = None


class LoadBinaryError(Exception):
    """
    An error loading the binary.
    """


class LoadBinary(QDialog):
    """
    Dialog displaying loading options for a binary.
    """

    def __init__(self, partial_ld, parent=None):
        super().__init__(parent)

        # initialization
        self.file_path = partial_ld.main_object.binary
        self.md5 = None
        self.sha256 = None
        self.option_widgets = {}
        self.is_blob = isinstance(partial_ld.main_object, Blob)
        self.arch = partial_ld.main_object.arch
        self.available_archs = archinfo.all_arches[::]

        if pypcode:
            for a in pypcode.Arch.enumerate():
                self.available_archs.extend(a.languages)

        # return values
        self.load_options = None

        self.setWindowTitle("Load a new binary")

        # checksums
        if hasattr(partial_ld.main_object, "md5") and partial_ld.main_object.md5 is not None:
            self.md5 = binascii.hexlify(partial_ld.main_object.md5).decode("ascii")
        if hasattr(partial_ld.main_object, "sha256") and partial_ld.main_object.sha256 is not None:
            self.sha256 = binascii.hexlify(partial_ld.main_object.sha256).decode("ascii")

        self.main_layout = QVBoxLayout()

        self._init_widgets()

        self._try_loading(partial_ld)

        self.setLayout(self.main_layout)

    @property
    def filename(self):
        return os.path.basename(self.file_path)

    #
    # Private methods
    #

    def _try_loading(self, partial_ld):
        deps = []
        processed_objects = set()
        for ident, obj in partial_ld._satisfied_deps.items():
            if obj is partial_ld._kernel_object or obj is partial_ld._extern_object or obj is partial_ld.main_object:
                continue
            if obj in processed_objects:
                continue
            deps.append(ident)
            processed_objects.add(obj)

        # dependencies

        dep_list: QListWidget = self.option_widgets["dep_list"]
        for dep in deps:
            dep_item = QListWidgetItem(dep)
            dep_item.setData(Qt.CheckStateRole, Qt.Unchecked)
            dep_list.addItem(dep_item)

    def _init_widgets(self):
        layout = QGridLayout()
        self.main_layout.addLayout(layout)

        # filename

        filename_caption = QLabel(self)
        filename_caption.setText("File name:")

        filename = QLabel(self)
        filename.setText(self.filename)

        layout.addWidget(filename_caption, 0, 0, Qt.AlignRight)
        layout.addWidget(filename, 0, 1)

        # md5

        if self.md5 is not None:
            md5_caption = QLabel(self)
            md5_caption.setText("MD5:")
            md5 = QLineEdit(self)
            md5.setText(self.md5)
            md5.setReadOnly(True)

            layout.addWidget(md5_caption, 1, 0, Qt.AlignRight)
            layout.addWidget(md5, 1, 1)

        # sha256

        if self.sha256 is not None:
            sha256_caption = QLabel(self)
            sha256_caption.setText("SHA256:")
            sha256 = QLineEdit(self)
            sha256.setText(self.sha256)
            sha256.setReadOnly(True)

            layout.addWidget(sha256_caption, 2, 0, Qt.AlignRight)
            layout.addWidget(sha256, 2, 1)

        # central tab

        tab = QTabWidget()
        self._init_central_tab(tab)

        self.main_layout.addWidget(tab)

        # buttons

        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self._on_cancel_clicked)
        self.main_layout.addWidget(buttons)

    def _init_central_tab(self, tab):
        self._init_load_options_tab(tab)

    def _init_load_options_tab(self, tab):
        arch_layout = QHBoxLayout()
        arch_caption = QLabel(self)
        arch_caption.setText("Architecture:")
        arch_caption.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        arch_layout.addWidget(arch_caption)
        arch_combo = QComboBox(self)

        index = 0
        for i, arch in enumerate(self.available_archs):
            if isinstance(arch, archinfo.Arch):
                arch_combo.addItem(f"{arch.bits}b {arch.name} ({arch.memory_endness[-2:]})", str(arch))
                if str(self.arch) == str(arch):
                    index = i
            elif pypcode and isinstance(arch, pypcode.ArchLanguage):
                arch_combo.addItem(f"{arch.id} (P-code Engine)")
                if self.arch.name == arch.id:
                    index = i
            else:
                assert False, "Unknown architecture type"

        arch_combo.setCurrentIndex(index)
        arch_layout.addWidget(arch_combo)
        self.option_widgets["arch"] = arch_combo

        if self.is_blob:
            blob_layout = QGridLayout()

            # load address
            base_addr_caption = QLabel(self)
            base_addr_caption.setText("Base Address:")
            blob_layout.addWidget(base_addr_caption, 1, 0)
            base_addr = QLineEdit(self)
            base_addr.setText("0")
            blob_layout.addWidget(base_addr, 1, 1)
            self.option_widgets["base_addr"] = base_addr

            # entry address
            entry_addr_caption = QLabel(self)
            entry_addr_caption.setText("Entry Address:")
            blob_layout.addWidget(entry_addr_caption, 2, 0)
            entry_addr = QLineEdit(self)
            entry_addr.setText("0")
            blob_layout.addWidget(entry_addr, 2, 1)
            self.option_widgets["entry_addr"] = entry_addr

        # load debug symbols
        load_debug_info = QCheckBox()
        load_debug_info.setText("Load debug information if available")
        load_debug_info.setChecked(True)
        self.option_widgets["load_debug_info"] = load_debug_info

        # auto load libs

        auto_load_libs = QCheckBox()
        auto_load_libs.setText("Automatically load all libraries (slow, not recommended)")
        auto_load_libs.setChecked(False)
        self.option_widgets["auto_load_libs"] = auto_load_libs

        # dependencies list

        dep_group = QGroupBox("Dependencies")
        dep_list = QListWidget()
        self.option_widgets["dep_list"] = dep_list

        sublayout = QVBoxLayout()
        sublayout.addWidget(dep_list)
        dep_group.setLayout(sublayout)

        layout = QVBoxLayout()
        if self.is_blob:
            layout.addLayout(blob_layout)
        layout.addLayout(arch_layout)
        layout.addWidget(load_debug_info)
        layout.addWidget(auto_load_libs)
        layout.addWidget(dep_group, stretch=1)

        frame = QFrame(self)
        frame.setLayout(layout)
        tab.addTab(frame, "Loading Options")

    #
    # Event handlers
    #

    def _on_ok_clicked(self):
        force_load_libs = []
        skip_libs = set()

        dep_list: QListWidget = self.option_widgets["dep_list"]
        for i in range(dep_list.count()):
            item: QListWidgetItem = dep_list.item(i)
            if item.checkState() == Qt.Checked:
                force_load_libs.append(item.text())
            else:
                skip_libs.add(item.text())

        self.load_options = {}
        self.load_options["auto_load_libs"] = self.option_widgets["auto_load_libs"].isChecked()
        self.load_options["load_debug_info"] = self.option_widgets["load_debug_info"].isChecked()

        arch = self.available_archs[self.option_widgets["arch"].currentIndex()]
        if pypcode and isinstance(arch, pypcode.ArchLanguage):
            arch = archinfo.ArchPcode(arch.id)
        self.load_options["arch"] = arch

        if self.is_blob:
            self.load_options["main_opts"] = {
                "backend": "blob",
                "base_addr": int(self.option_widgets["base_addr"].text(), 16),
                "entry_point": int(self.option_widgets["entry_addr"].text(), 16),
            }

        if force_load_libs:
            self.load_options["force_load_libs"] = force_load_libs
        if skip_libs:
            self.load_options["skip_libs"] = skip_libs

        self.close()

    def _on_cancel_clicked(self):
        self.close()

    @staticmethod
    def run(partial_ld) -> Tuple[Optional[Dict], Optional[Dict], Optional[Dict]]:
        try:
            dialog = LoadBinary(partial_ld)
            dialog.setModal(True)
            dialog.exec_()
            return dialog.load_options
        except LoadBinaryError:
            pass
        return None, None, None

    @staticmethod
    def binary_arch_detect_failed(filename: str, archinfo_msg: str):
        # TODO: Normalize the path for Windows
        QMessageBox.warning(None, "Architecture selection failed", f"{archinfo_msg} for binary:\n\n{filename}")

    @staticmethod
    def binary_loading_failed(filename):
        # TODO: Normalize the path for Windows
        QMessageBox.critical(None, "Failed to load binary", f"angr failed to load binary {filename}.")



from PySide6.QtWidgets import QFileDialog, QHBoxLayout, QPushButton, QVBoxLayout

from angrmanagement.ui.widgets.qpatch_table import QPatchTable

from .view import BaseView


class PatchesView(BaseView):
    def __init__(self, instance, default_docking_position, *args, **kwargs):
        super().__init__("patches", instance, default_docking_position, *args, **kwargs)

        self.base_caption = "Patches"
        self._patch_table: QPatchTable

        self._init_widgets()

    def reload(self):
        self._patch_table.reload()

    #
    # Private methods
    #

    def _init_widgets(self):
        # patch table
        self._patch_table = QPatchTable(self.instance, self)

        #
        # controls
        #
        control_layout = QHBoxLayout()

        # apply patches
        btn_apply_patches = QPushButton("Save patched binary as...")
        btn_apply_patches.clicked.connect(self._on_apply_patches_clicked)
        control_layout.addStretch(0)
        control_layout.addWidget(btn_apply_patches)
        control_layout.setContentsMargins(3, 3, 3, 3)

        layout = QVBoxLayout()
        layout.addLayout(control_layout)
        layout.addWidget(self._patch_table)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    #
    # Events
    #

    def _on_apply_patches_clicked(self):
        # select where the new binary will be stored
        # Open File window
        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Save the patched binary to...",
            self.instance.project.loader.main_object.binary
            + ".patched",  # FIXME: this will not work if we are loading from an angrdb
            "Any file (*)",
        )

        if file_path:
            b = self.instance.project.kb.patches.apply_patches_to_binary()
            with open(file_path, "wb") as f:
                f.write(b)



import asyncio
import math
import threading
import time
from time import sleep
from typing import Optional

from PySide6.QtCore import QEvent, QObject
from PySide6.QtWidgets import QDialog, QLabel, QLineEdit, QPushButton, QVBoxLayout

from angrmanagement.plugins.base_plugin import BasePlugin

try:
    from slacrs import Slacrs
    from slacrs.model import HumanFatigue
except ImportError:
    Slacrs: Optional[type] = None
    HumanFatigue: Optional[type] = None
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

#
# Plugin to capture the User Mouse Movements.
# User must input their name to use this plugin.
#


class LogFatiguePlugin(BasePlugin):
    def __init__(self, workspace):
        if not Slacrs:
            raise Exception("Please install Slacrs to Initialize LogFatigue Plugin")
        self._fatigue_flag = True
        super().__init__(workspace)
        self._fatigue = HumanFatigue()
        self._strokes = []
        self._main_window = workspace.view_manager.main_window
        self._main_window.setMouseTracking(True)
        self.EventFilterInstance = self.EventFilter(self._fatigue)
        self._main_window.installEventFilter(self.EventFilterInstance)
        self.modal = self.Modal(self)
        self.modal.show()

        self.t_log = threading.Thread(target=self._log_mouse, args=())
        self.t_log.setDaemon(True)
        self.t_log.start()

    def _log_mouse(self):
        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())

        connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
        if connector is None:
            # chess connector does not exist
            return None
        slacrs_instance = connector.slacrs_instance()
        if slacrs_instance is None:
            # slacrs does not exist. continue
            return None
        session = slacrs_instance.session()
        with session.no_autoflush:
            while self._fatigue_flag is True:
                sleep(2)
                if self._fatigue.user:
                    session.add(self._fatigue)
                    session.commit()
            session.close()

    class EventFilter(QObject):
        def __init__(self, fatigue):
            super().__init__()
            self._fatigue = fatigue
            self._strokes = []

        def eventFilter(self, obj, event):  # pylint: disable=unused-argument
            if event.type() == QEvent.HoverMove:
                x = event.pos().x()
                old_x = event.oldPos().x()
                y = event.pos().y()
                old_y = event.oldPos().y()

                self._fatigue.mouse_speed = int(math.sqrt((x - old_x) ** 2 + (y - old_y) ** 2))
            elif event.type() == QEvent.KeyPress:
                timestamp = time.time()
                i = 0
                for i in range(len(self._strokes)):
                    if timestamp - self._strokes[i] <= 10:
                        break
                self._strokes = self._strokes[i:]
                self._strokes.append(timestamp)
                self._fatigue.stroke = len(self._strokes)

            return False

    #
    # Creates Model so user can input name, if modal is closed without submitting name,
    # the LogFatigue plugin will be deactivated
    #

    class Modal(QDialog):
        def __init__(self, outerclass):
            super().__init__(outerclass._main_window)

            self.label = QLabel("Enter your Name")
            self.edit = QLineEdit("")
            self.button = QPushButton("Submit")
            self.outerclass = outerclass

            layout = QVBoxLayout()

            layout.addWidget(self.label)
            layout.addWidget(self.edit)
            layout.addWidget(self.button)

            self.setLayout(layout)

            self.button.clicked.connect(self.input)

        def input(self):
            self.outerclass._fatigue.user = self.edit.text()
            self.close()

        def closeEvent(self, event):
            if not self.outerclass._fatigue.user:
                self.outerclass._fatigue_flag = False
                self.outerclass.workspace.plugins.deactivate_plugin(self.outerclass)
            event.accept()

    def teardown(self):
        self._fatigue_flag = False
        self.t_log.join()



import logging
import random
from collections import defaultdict

from angr.errors import SimEngineError
from PySide6.QtGui import QColor

log = logging.getLogger(__name__)


class TraceFunc:
    """
    Trace and Function class

    """

    def __init__(self, bbl_addr=None, func_name=None, func=None):
        self.bbl_addr = bbl_addr
        self.func_name = func_name
        self.func = func


class TraceStatistics:
    """
    Trace and color / legend record
    """

    BBL_FILL_COLOR = QColor(0, 0xF0, 0xF0, 0xF)
    BBL_BORDER_COLOR = QColor(0, 0xF0, 0xF0)
    BBL_EMPTY_COLOR = QColor("white")

    def __init__(self, workspace, trace, trace_id=None, baddr=None):
        self.workspace = workspace
        self.bbl_addrs = trace
        self.id = trace_id
        self.trace_func = []
        self._func_color = {}
        self.count = None
        self._mark_color = {}
        self._positions = defaultdict(list)
        self.mapped_trace = []

        project = self.workspace.main_instance.project
        if project.am_none:
            self.project_baddr = None
        else:
            self.project_baddr = project.loader.main_object.mapped_base
        if baddr is None:
            self.runtime_baddr = self.project_baddr
        else:
            self.runtime_baddr = baddr

        self._statistics(self.bbl_addrs)

    def get_func_color(self, func_name):
        if func_name in self._func_color:
            return self._func_color[func_name]
        else:
            color = self._random_color()
            self._func_color[func_name] = color
        return color

    def set_mark_color(self, p, color):
        self._mark_color[p] = color

    def get_mark_color(self, addr, i):
        try:
            mark_index = self._get_position(addr, i)
            mark_color = self._mark_color[mark_index]
        except (IndexError, KeyError) as e:
            log.error(e)
            return self.BBL_EMPTY_COLOR
        return mark_color

    def get_positions(self, addr):
        return self._positions[addr]

    def get_count(self, ins):
        return len(self._positions[ins])

    def get_bbl_from_position(self, position):
        return self.trace_func[position].bbl_addr

    def get_func_name_from_position(self, position):
        return self.trace_func[position].func_name

    def get_func_from_position(self, position):
        return self.trace_func[position].func

    def _apply_trace_offset(self, addr):
        if self.project_baddr is None:
            return None
        offset = self.project_baddr - self.runtime_baddr
        return addr + offset

    def _statistics(self, trace_addrs):
        """
        :param trace: basic block address list
        """
        self.mapped_trace = []
        for addr in trace_addrs:
            converted = self._apply_trace_offset(addr)
            if converted is not None:
                self.mapped_trace.append(converted)

        bbls = filter(self._get_bbl, self.mapped_trace)

        for p, bbl_addr in enumerate(bbls):
            block = self.workspace.main_instance.project.factory.block(bbl_addr)
            for addr in block.instruction_addrs:
                self._positions[addr].append(p)

            node = self.workspace.main_instance.cfg.get_any_node(bbl_addr)
            if node is None:  # try again without asssuming node is start of a basic block
                node = self.workspace.main_instance.cfg.get_any_node(bbl_addr, anyaddr=True)

            func_name = hex(bbl_addr)  # default to using bbl_addr as name if none is not found
            func = None
            if node is not None:
                func_addr = node.function_address
                functions = self.workspace.main_instance.project.kb.functions
                if func_addr in functions:
                    func_name = functions[func_addr].demangled_name
                    func = functions[func_addr]
                else:
                    func_name = "Unknown"
            else:
                log.warning("Node at %x is None, using bbl_addr as function name", bbl_addr)
            self.trace_func.append(TraceFunc(bbl_addr, func_name, func))

        self.count = len(self.trace_func)

    def _get_bbl(self, addr):
        try:
            return self.workspace.main_instance.project.factory.block(addr)
        except SimEngineError:
            return None

    def _func_addr(self, a):
        return self.workspace.main_instance.cfg.get_any_node(a).function_address

    def _func_name(self, a):
        return self.workspace.main_instance.project.kb.functions[self._func_addr(a)].demangled_name

    @staticmethod
    def _random_color():
        r = random.randint(0, 255)
        g = random.randint(0, 255)
        b = random.randint(0, 255)
        return QColor(r, g, b)

    def _get_position(self, addr, i):
        return self._positions[addr][i]



# pylint:disable=unused-argument
import asyncio
import threading
import typing
from typing import TYPE_CHECKING, List, Optional, Tuple

from PySide6.QtCore import QAbstractTableModel, Qt
from PySide6.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QMessageBox,
    QPushButton,
    QTableView,
    QVBoxLayout,
)
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from angrmanagement.logic.threads import gui_thread_schedule_async

if TYPE_CHECKING:
    import PySide6

    from angrmanagement.ui.workspace import Workspace

try:
    import slacrs
except ImportError:
    slacrs = None


class TraceDescriptor:
    """
    Models a trace.
    """

    def __init__(self, trace_id: str, input_id: str, created_at, type_: str):
        self.trace_id = trace_id
        self.input_id = input_id
        self.created_at = created_at
        self.type = type_


class QTraceTableModel(QAbstractTableModel):
    """
    Implements a table model for traces.
    """

    Headers = ["Trace ID", "Created at", "Input ID", "Input Length", "Type"]
    COL_TRACEID = 0
    COL_CREATEDAT = 1
    COL_INPUTID = 2
    COL_INPUTLENGTH = 3
    COL_TYPE = 4

    def __init__(self):
        super().__init__()
        self._traces: List[TraceDescriptor] = []

    @property
    def traces(self):
        return self._traces

    @traces.setter
    def traces(self, v):
        self.beginResetModel()
        self._traces = v
        self.endResetModel()

    def rowCount(self, parent: "PySide6.QtCore.QModelIndex" = ...) -> int:
        return len(self.traces)

    def columnCount(self, parent: "PySide6.QtCore.QModelIndex" = ...) -> int:
        return len(self.Headers)

    def headerData(self, section: int, orientation: "PySide6.QtCore.Qt.Orientation", role: int = ...) -> typing.Any:
        if role != Qt.DisplayRole:
            return None

        if section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index: "PySide6.QtCore.QModelIndex", role: int = ...) -> typing.Any:
        if not index.isValid():
            return None
        row = index.row()
        if row >= len(self.traces):
            return None
        trace = self.traces[row]
        col = index.column()

        if role == Qt.DisplayRole:
            return self._get_column_text(trace, col)

        return None

    @staticmethod
    def _get_column_text(trace: TraceDescriptor, col: int) -> str:
        mapping = {
            QTraceTableModel.COL_TRACEID: QTraceTableModel._get_trace_id,
            QTraceTableModel.COL_CREATEDAT: QTraceTableModel._get_trace_created_at,
            QTraceTableModel.COL_TYPE: QTraceTableModel._get_trace_type,
            QTraceTableModel.COL_INPUTID: QTraceTableModel._get_trace_input_id,
            QTraceTableModel.COL_INPUTLENGTH: lambda x: "Unknown",
        }
        return mapping[col](trace)

    @staticmethod
    def _get_trace_id(trace: TraceDescriptor) -> str:
        return trace.trace_id

    @staticmethod
    def _get_trace_created_at(trace: TraceDescriptor) -> str:
        return trace.created_at

    @staticmethod
    def _get_trace_type(trace: TraceDescriptor) -> str:
        return trace.type

    @staticmethod
    def _get_trace_input_id(trace: TraceDescriptor) -> str:
        return trace.input_id


class QTraceTableView(QTableView):
    """
    Implements a trace view for CHESS traces.
    """

    def __init__(self):
        super().__init__()

        self.horizontalHeader().setVisible(True)
        self.verticalHeader().setVisible(False)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.MultiSelection)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)

        self.model: QTraceTableModel = QTraceTableModel()
        self.setModel(self.model)


class QChessTraceListDialog(QDialog):
    """
    Implements a CHESS trace list dialog.
    """

    def __init__(self, workspace: "Workspace", parent=None):
        super().__init__(parent)

        if slacrs is None:
            QMessageBox.Critical(
                self,
                "Slacrs is not installed",
                "Cannot import slacrs. Please make sure slacrs is properly installed.",
                QMessageBox.Ok,
            )
            self.close()
            return

        self.workspace = workspace
        self.trace_ids: Optional[List[Tuple[str, str]]] = None  # input ID, trace ID
        self.setMinimumWidth(400)

        self._status_label: QLabel = None
        self._table: QTraceTableView = None
        self._ok_button: QPushButton = None
        self._cancel_button: QPushButton = None

        self.setWindowTitle("Open traces from CHECRS")
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        self._init_widgets()

        self._status_label.setText("Loading...")
        self.workspace.main_window.app.processEvents()
        th = threading.Thread(target=self._load_traces, daemon=True)
        th.start()

    def _init_widgets(self):
        # table
        self._table = QTraceTableView()

        # status
        status_lbl = QLabel("Status:")
        self._status_label = QLabel()
        status_layout = QHBoxLayout()
        status_layout.addWidget(status_lbl)
        status_layout.addWidget(self._status_label)
        status_layout.addStretch(0)

        # buttons
        self._ok_button = QPushButton("Ok")
        self._ok_button.clicked.connect(self._on_ok_button_clicked)
        self._cancel_button = QPushButton("Cancel")
        self._cancel_button.clicked.connect(self._on_cancel_button_clicked)
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self._ok_button)
        buttons_layout.addWidget(self._cancel_button)

        layout = QVBoxLayout()
        layout.addWidget(self._table)
        layout.addLayout(status_layout)
        layout.addLayout(buttons_layout)
        self.setLayout(layout)

    def _load_traces(self):
        from slacrs.model import Input, Trace  # pylint:disable=import-outside-toplevel,import-error

        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())

        connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
        if connector is None:
            return

        connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
        if connector is None:
            # chess connector does not exist
            return
        slacrs_instance = connector.slacrs_instance()
        if slacrs_instance is None:
            # slacrs does not exist. continue
            return

        session = slacrs_instance.session()
        target_image_id = connector.target_image_id
        if not target_image_id:
            return

        traces: List[TraceDescriptor] = []

        db_traces = session.query(Trace).join(Trace.input).filter(Input.target_image_id == target_image_id)
        for db_trace in db_traces:
            db_trace: Trace
            t = TraceDescriptor(db_trace.id, db_trace.input_id, db_trace.created_at, "block trace")
            traces.append(t)

        session.close()
        gui_thread_schedule_async(self._update_table, args=(traces,))

    def _update_table(self, traces):
        self._table.model.traces = traces
        self._table.viewport().update()
        self._status_label.setText("Ready.")

    #
    # Events
    #

    def _on_ok_button_clicked(self):
        selection_model = self._table.selectionModel()
        if not selection_model.hasSelection():
            QMessageBox.warning(
                self, "No target is selected", "Please select a CHESS target to continue.", QMessageBox.Ok
            )
            return

        rows = selection_model.selectedRows()
        self.trace_ids = []
        for row in rows:
            trace = self._table.model.traces[row.row()]
            self.trace_ids.append((trace.input_id, trace.trace_id))
        self.close()

    def _on_cancel_button_clicked(self):
        self.close()



from .job import Job


class CodeTaggingJob(Job):
    def __init__(self, on_finish=None):
        super().__init__(name="Code tagging", on_finish=on_finish)

    def _run(self, inst):
        func_count = len(inst.kb.functions)
        for i, func in enumerate(inst.kb.functions.values()):
            if func.alignment:
                continue
            ct = inst.project.analyses.CodeTagging(func)
            func.tags = tuple(ct.tags)

            percentage = i / func_count * 100
            text = "%.02f%%" % percentage

            super()._progress_callback(percentage, text=text)

    def finish(self, inst, result):
        super().finish(inst, result)

    def __repr__(self):
        return "CodeTaggingJob"



# pylint:disable=global-statement,missing-class-docstring,no-self-use,unspecified-encoding
import logging
import os
import subprocess
import sys
from pathlib import Path
from typing import Optional, Tuple

import tomlkit
import tomlkit.exceptions
from PySide6.QtWidgets import QApplication, QFileDialog, QMessageBox
from xdg import BaseDirectory

from angrmanagement.daemon.server import register_server_exposed_method
from angrmanagement.daemon.url_handler import UrlActionBase, register_url_action
from angrmanagement.plugins import BasePlugin

_l = logging.getLogger(name=__name__)

# we probably want to put this feature into angr management
_app = None


def tmp_app():
    global _app
    if _app is None:
        _app = QApplication()
    return _app


class UrlActionOpenSourceFile(UrlActionBase):
    def __init__(
        self, target_uuid: str, challenge_name: str, source_file: str, line_number: str, position: str, editor: str
    ):
        super().__init__()

        self.target_uuid = target_uuid
        self.challenge_name = challenge_name
        self.source_file = source_file
        self.editor = editor

        # TODO: Parse line number and position
        self.line_number = line_number
        self.position = position

    def act(self, daemon_conn=None):
        daemon_conn.root.open_source_file(
            self.target_uuid,
            self.challenge_name,
            self.source_file,
            self.line_number,
            self.position,
            self.editor,
        )

    @classmethod
    def _from_params(cls, params):
        return cls(
            cls._one_param(params, "target_uuid"),
            cls._one_param(params, "challenge_name"),
            cls._one_param(params, "source_file"),
            cls._one_param(params, "line_number"),
            cls._one_param(params, "position"),
            cls._one_param(params, "editor"),
        )


class ChessUrlHandler(BasePlugin):
    DISPLAY_NAME = "CHESS URL Handler"
    REQUIRE_WORKSPACE = False

    def __init__(self, workspace):
        super().__init__(workspace)

        self._register_url_handlers()

    def _register_url_handlers(self):
        register_url_action("open_source_file", UrlActionOpenSourceFile)
        register_server_exposed_method("open_source_file", self.exposed_open_source_file)

    def _get_rootdir_config_path(self) -> str:
        am_root = BaseDirectory.save_config_path("angr-management")
        if am_root is not None:
            rootdirs_path = os.path.join(am_root, "challenge_rootdirs.toml")
            return rootdirs_path
        raise ValueError("Cannot get the configuration file root directory for angr-management.")

    def _get_rootdir(
        self, target_uuid: str, challenge_name: str, source_file: str
    ) -> Tuple[Optional[str], Optional[str]]:
        rootdirs_path = self._get_rootdir_config_path()
        # load it if it exists
        entries = {}
        if os.path.isfile(rootdirs_path):
            with open(rootdirs_path) as f:
                try:
                    entries = tomlkit.load(f)
                except tomlkit.exceptions.ParseError:
                    _l.error("Cannot decode rootdirs file %s. Ignore existing content.", rootdirs_path)

        dir_path = None
        if "uuid_to_rootdir" in entries:
            if target_uuid in entries["uuid_to_rootdir"]:
                # yes there is already one
                dir_path = entries["uuid_to_rootdir"][target_uuid].value
            else:
                dir_path = None

        while True:
            # test the existing dir_path for the existence of the file
            if dir_path:
                if not source_file:
                    # source file is not specified. we take whatever is there!
                    return dir_path, None

                file_found = False
                full_path = None

                source_file_name = os.path.basename(source_file)
                for base, _, files in os.walk(dir_path):
                    if source_file_name in files:
                        # check the full path
                        full_path = os.path.normpath(os.path.join(base, source_file_name))
                        if full_path.endswith(source_file):
                            # found it!
                            file_found = True
                            break

                if file_found:
                    return dir_path, full_path

            # we either did not find the file under the specified directory or got a wrong directory to start with
            # ask the user to manually specify a directory
            tmp_app()
            QMessageBox.information(
                None,
                "Specifying the root directory",
                f"File {source_file} is not found under the directory you specified.\n\n"
                f"You will be asked to specify the challenge root directory for target {target_uuid} "
                f"({challenge_name}). We will then open the specified source code file for you.\n"
                f"We will only ask you once per target ID. Your selection will be stored on this machine.",
            )
            dir_path = QFileDialog.getExistingDirectory(
                None,
                f"Specifying the root directory for target {target_uuid} ({challenge_name})",
                # TODO: Use the default challenge root directory
                str(Path.home()),
                QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks,
            )
            if not dir_path:
                # the user did not specify anything
                return None, None

            # loop back to test the validity of the provided root directory

    def _save_rootdir(self, target_uuid: str, challenge_name: str, root_dir: str):
        rootdirs_path = self._get_rootdir_config_path()
        # load it
        entries = {}
        if os.path.isfile(rootdirs_path):
            with open(rootdirs_path) as f:
                try:
                    entries = tomlkit.load(f)
                except tomlkit.exceptions.ParseError:
                    _l.error("Cannot decode rootdirs file %s. Ignore existing content.", rootdirs_path)

        if "uuid_to_challenge" not in entries:
            entries["uuid_to_challenge"] = {}
        if "uuid_to_rootdir" not in entries:
            entries["uuid_to_rootdir"] = {}
        entries["uuid_to_challenge"][target_uuid] = challenge_name
        entries["uuid_to_rootdir"][target_uuid] = root_dir

        # store it
        with open(rootdirs_path, "w") as f:
            tomlkit.dump(entries, f)

    def _vscode_path(self) -> Optional[str]:
        if sys.platform == "win32":
            default_locations = [
                os.path.join(os.getenv("LOCALAPPDATA"), "Programs", "Microsoft VS Code"),
            ]
            default_locations += os.getenv("PATH").split(";")
            code_exe = "code.exe"
        elif sys.platform == "linux":
            # assuming code is within PATH
            return "code"
        elif sys.platform == "darwin":
            # try the default path
            raise NotImplementedError()
        else:
            raise NotImplementedError()

        for loc in default_locations:
            vscode_path = os.path.join(loc, code_exe)
            if os.path.isfile(vscode_path):
                # found it!
                if sys.platform in ("linux", "darwin"):
                    return vscode_path

        return None

    def exposed_open_source_file(
        self, target_uuid: str, challenge_name: str, source_file: str, line_number: int, position: int, editor: str
    ) -> None:
        # Find the root directory of the challenge
        # if this is the first time we see a challenge name, ask the user to specify the challenge root directory
        root_dir, file_path = self._get_rootdir(target_uuid, challenge_name, source_file)

        if not root_dir:
            return

        # log the root dir for the given target UUID and challenge_name
        self._save_rootdir(target_uuid, challenge_name, root_dir)

        if not file_path:
            return

        # invoke the corresponding editor
        if editor == "vscode":
            # https://code.visualstudio.com/docs/editor/command-line#_opening-files-and-folders
            vscode = self._vscode_path()
            if vscode:
                cmd_line = [vscode, "-g", f"{file_path}:{line_number}:{position}"]
                subprocess.Popen(cmd_line, close_fds=True)
            else:
                tmp_app()
                QMessageBox.critical(None, "VSCode is not found", "Cannot find the executable for VS Code.")
                return

        elif editor == "am":
            raise NotImplementedError()

        else:
            # fallback to the default editor
            raise NotImplementedError()



from typing import Optional

from angrmanagement.config import Conf
from angrmanagement.plugins.base_plugin import BasePlugin

try:
    from slacrs import Slacrs
    from slacrs.model import FunctionRename, ReverseEngineeringProgress, VariableRename
except ImportError:
    Slacrs: Optional[type] = None
    VariableRename: Optional[type] = None
    FunctionRename: Optional[type] = None
    ReverseEngineeringProgress: Optional[type] = None


class LogReverseEngineeringPlugin(BasePlugin):
    """
    Plugin for logging the reverse engineering of a program
    """

    def __init__(self, workspace):
        if not Slacrs:
            raise Exception("Please install Slacrs to Initialize LogReverseEngineering Plugin")
        super().__init__(workspace)
        self.session = Slacrs(database=Conf.checrs_backend_str).session()
        self.project = (
            self.workspace.main_instance.img_name
            if self.workspace.main_instance.img_name
            else self.workspace.main_instance.project.filename
        )

    def handle_stack_var_renamed(self, func, offset, old_name, new_name):
        """
        Logic to check if the same variable has already been renamed, if not add to the current session.
        """
        if offset:
            new_name = old_name
        variable_rename = (
            self.session.query(VariableRename)
            .filter(
                VariableRename.project == self.project,
                VariableRename.function == func._name,
                VariableRename.variable == old_name,
            )
            .first()
        )
        if variable_rename:
            self.session.delete(variable_rename)
        variable_rename = VariableRename()
        variable_rename.project = self.project
        variable_rename.function = func._name
        variable_rename.variable = new_name
        self.session.add(variable_rename)

    def handle_function_renamed(self, func, old_name: str, new_name: str):
        """
        Logic to check if the same Function has already been renamed, if not add to the current session.
        """
        function_rename = (
            self.session.query(FunctionRename)
            .filter(
                FunctionRename.project == self.project,
                FunctionRename.function == old_name,
            )
            .first()
        )
        if old_name.startswith("sub") or function_rename:
            if function_rename:
                self.session.delete(function_rename)
            function_rename = FunctionRename()
            function_rename.project = self.project
            function_rename.function = new_name
            self.update_function_name(old_name, new_name)
            self.session.add(function_rename)

    def handle_project_save(self, file_name: str):
        """
        Commit the current session only when user saves the project, uncommitted session objects will be discarded
        at teardown.
        """
        variables_renamed_count = len(
            self.session.query(VariableRename).filter(VariableRename.project == self.project).all()
        )
        total_variables_count = len(self.workspace.main_instance.project.kb.variables.global_manager._variables)
        reverse_eng_progress = (
            self.session.query(ReverseEngineeringProgress)
            .filter(ReverseEngineeringProgress.project == self.project)
            .first()
        )
        if not reverse_eng_progress:
            reverse_eng_progress = ReverseEngineeringProgress()
            self.session.add(reverse_eng_progress)
        reverse_eng_progress.project = self.project
        reverse_eng_progress.variables_renamed = variables_renamed_count
        reverse_eng_progress.total_variables = total_variables_count
        (
            reverse_eng_progress.functions_renamed,
            reverse_eng_progress.total_functions,
        ) = self.get_function_rename_stats()
        self.session.commit()

    def update_function_name(self, old_name, new_name):
        """
        To update the function names for all variable_rename if function gets renamed.
        """
        variables_renamed = self.session.query(VariableRename).filter(
            VariableRename.project == self.project, VariableRename.function == old_name
        )
        for obj in variables_renamed:
            obj.function = new_name

    def get_function_rename_stats(self):
        functions_renamed = [
            func.function
            for func in self.session.query(FunctionRename).filter(FunctionRename.project == self.project).all()
        ]

        functions_renamed_count = 0
        total_functions_count = 0

        for key in self.workspace.main_instance.project.kb.functions._function_map:
            if self.workspace.main_instance.project.kb.functions._function_map[key]._name in functions_renamed:
                functions_renamed_count = functions_renamed_count + 1
                total_functions_count = total_functions_count + 1
            elif self.workspace.main_instance.project.kb.functions._function_map[key]._name.startswith("sub"):
                total_functions_count = total_functions_count + 1

        return [functions_renamed_count, total_functions_count]

    def teardown(self):
        self.session.close()



from PySide6.QtCore import QSize
from PySide6.QtWidgets import QHBoxLayout

from angrmanagement.ui.widgets.qstate_table import QStateTable

from .view import BaseView


class StatesView(BaseView):
    def __init__(self, instance, default_docking_position, *args, **kwargs):
        super().__init__("states", instance, default_docking_position, *args, **kwargs)

        self.base_caption = "States"
        self._state_table: QStateTable

        self._init_widgets()

    def reload(self):
        self._state_table.state_manager = self.instance.states

    def closeEvent(self, event):
        """
        Close children before exiting
        """
        self._state_table.close()

    def sizeHint(self):
        return QSize(400, 800)

    def _init_widgets(self):
        self._state_table = QStateTable(self.instance, self, selection_callback=self._on_state_selected)

        hlayout = QHBoxLayout()
        hlayout.addWidget(self._state_table)
        hlayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(hlayout)

    def _on_state_selected(self, state):
        """
        A new function is on selection right now. Update the disassembly view that is currently at front.

        :param function:
        :return:
        """

        pass



from typing import List, Optional


class ComponentFunction:
    __slots__ = (
        "mapped_base",
        "virtual_addr",
        "symbol_name",
    )

    def __init__(self, mapped_base: int, virtual_addr: int, symbol_name: Optional[str] = None):
        self.mapped_base = mapped_base
        self.virtual_addr = virtual_addr
        self.symbol_name = symbol_name

    def __eq__(self, other):
        return (
            isinstance(other, ComponentFunction)
            and self.mapped_base == other.mapped_base
            and self.virtual_addr == other.virtual_addr
        )

    def __hash__(self):
        return hash((ComponentFunction, self.mapped_base, self.virtual_addr))


class ComponentTreeNode:
    def __init__(self, name=None):
        self.name = name
        self.components: List["ComponentTreeNode"] = []
        self.functions: List[ComponentFunction] = []

    def __eq__(self, other):
        return (
            isinstance(other, ComponentTreeNode)
            and self.components == other.components
            and set(self.functions) == set(other.functions)
        )

    def __hash__(self):
        return hash(
            (
                ComponentTreeNode,
                hash(tuple(self.components)),
                hash(tuple(sorted((f.mapped_base + f.virtual_addr) for f in self.functions))),
            )
        )


class ComponentTree:
    def __init__(self, root: Optional[ComponentTreeNode] = None):
        self.root = root



# pylint:disable=ungrouped-imports
import os
import threading
import time
from typing import TYPE_CHECKING, Optional

from PySide6.QtGui import QIcon, QPixmap, Qt
from PySide6.QtWidgets import QMessageBox, QPushButton

from angrmanagement.config import Conf, save_config
from angrmanagement.config.config_entry import ConfigurationEntry
from angrmanagement.daemon.client import DaemonClient
from angrmanagement.logic.threads import gui_thread_schedule_async
from angrmanagement.plugins import BasePlugin

from .backend_selector_dialog import QBackendSelectorDialog
from .summary_view import SummaryView
from .target_selector import QTargetSelectorDialog

try:
    import slacrs
except ImportError:
    slacrs = None

if TYPE_CHECKING:
    from slacrs import Slacrs

    from angrmanagement.ui.workspace import Workspace


class ChessConnector(BasePlugin):
    """
    Implements the CHESS connector plugin.
    """

    DISPLAY_NAME = "CHESS Connector"
    REQUIRE_WORKSPACE = True

    def __init__(self, workspace):
        super().__init__(workspace)

        image_file_root = os.path.dirname(__file__)

        status_bar_height = int(workspace.main_window.statusBar().height() * 0.75)
        self._connected_pixmap = QPixmap(os.path.join(image_file_root, "connected.png")).scaled(
            status_bar_height, status_bar_height, Qt.AspectRatioMode.IgnoreAspectRatio, Qt.SmoothTransformation
        )
        self._connecting_pixmap = QPixmap(os.path.join(image_file_root, "connecting.png")).scaled(
            status_bar_height, status_bar_height, Qt.AspectRatioMode.IgnoreAspectRatio, Qt.SmoothTransformation
        )
        self._disconnected_pixmap = QPixmap(os.path.join(image_file_root, "disconnected.png")).scaled(
            status_bar_height, status_bar_height, Qt.AspectRatioMode.IgnoreAspectRatio, Qt.SmoothTransformation
        )

        self._status_button = QPushButton()
        self._status_button.setFlat(True)
        self._status_button.clicked.connect(self._on_status_button_clicked)

        self.target_id: Optional[str] = None
        self.target_image_id: Optional[str] = None
        self.target_description: Optional[str] = None
        self._target_description_label = QPushButton()
        self._target_description_label.setFlat(True)
        self._target_description_label.clicked.connect(self.set_chess_target)
        self.target_id_updated()

        self._slacrs_instance: Optional["Slacrs"] = None
        self._slacrs_database_str: Optional[str] = None

        self.connected: bool = False
        self.backend_disconnected()

        self.active = True

        th = threading.Thread(
            target=self.checrs_connection_monitor,
            args=(),
        )
        th.setDaemon(True)
        th.start()

        self.summary_view = SummaryView(self.workspace, "center", self)
        self.workspace.add_view(self.summary_view)

    def teardown(self):
        self.active = False
        self.summary_view.teardown()

    def slacrs_instance(self, database: Optional[str] = None):
        if not database:
            # load the default database str
            database = Conf.checrs_backend_str

        # again, if database_str is empty, return nothing
        if not database:
            return None

        if database == self._slacrs_database_str and self._slacrs_instance is not None:
            return self._slacrs_instance

        self._slacrs_instance = slacrs.Slacrs(database=database)
        # if it works out, remember the database str
        self._slacrs_database_str = database

        return self._slacrs_instance

    def backend_disconnected(self):
        self.connected = False
        self._status_button.setIcon(QIcon(self._disconnected_pixmap))
        self._status_button.setToolTip("Disconnected from CHECRS backend. Retry every 5 seconds...")
        self.workspace.main_window.app.processEvents()

    def backend_connected(self):
        self.connected = True
        self._status_button.setIcon(QIcon(self._connected_pixmap))
        self._status_button.setToolTip("Connected to CHECRS backend")
        self.workspace.main_window.app.processEvents()

    def backend_connecting(self):
        if not self.connected:
            # only change the icon to connecting if we are previously not connected
            self._status_button.setIcon(QIcon(self._connecting_pixmap))
            self._status_button.setToolTip("Connecting to CHECRS backend")
            self.workspace.main_window.app.processEvents()

    def target_id_updated(self):
        if self.target_id and self.target_image_id:
            desc = self.target_description if self.target_description else f"(no description) {self.target_id}"
            self._target_description_label.setText(desc)
            self._target_description_label.setToolTip(
                f"Target ID: {self.target_id}\n" f"Target image ID: {self.target_image_id}"
            )

            DaemonClient.register_binary(self.workspace.main_instance.project.loader.main_object.binary, self.target_id)
        else:
            self._target_description_label.setText("No associated CHESS target")
            self._target_description_label.setToolTip("")

    def status_bar_permanent_widgets(self):
        yield self._target_description_label
        yield self._status_button

    def on_workspace_initialized(self, workspace: "Workspace"):
        pass

    def _on_status_button_clicked(self):
        self.set_checrs_backend_str()

    def checrs_connection_monitor(self):
        first_run = True
        while self.active:
            if first_run:
                first_run = False
            else:
                time.sleep(5)

            if slacrs is None or not hasattr(Conf, "checrs_backend_str") or not Conf.checrs_backend_str:
                gui_thread_schedule_async(self.backend_disconnected)
                continue

            engine = None
            try:
                gui_thread_schedule_async(self.backend_connecting)
                engine = slacrs.Slacrs.connect_to_db(Conf.checrs_backend_str)
                connection = engine.connect()
                connection.close()  # close it immediately
            except Exception:  # pylint:disable=broad-except
                gui_thread_schedule_async(self.backend_disconnected)
                continue
            finally:
                if engine is not None:
                    engine.dispose()

            gui_thread_schedule_async(self.backend_connected)

    #
    # Custom menu actions
    #

    MENU_BUTTONS = ["Connect to CHECRS backend...", "Set associated CHESS target..."]
    CONNECT_TO_BACKEND = 0
    SET_CHESS_TARGET = 1

    def handle_click_menu(self, idx):
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return
        mapping = {
            ChessConnector.CONNECT_TO_BACKEND: self.set_checrs_backend_str,
            ChessConnector.SET_CHESS_TARGET: self.set_chess_target,
        }
        mapping[idx]()

    def set_checrs_backend_str(self):
        dialog = QBackendSelectorDialog(
            self.workspace,
            backend_str=Conf.checrs_backend_str,
            rest_backend_str=Conf.checrs_rest_endpoint_url,
            chess_connector=self,
            parent=self.workspace.main_window,
        )
        dialog.exec_()

        server_url = dialog.backend_str
        if server_url is not None:
            # if it's empty, it means the user wants to disconnect from CHECRS
            Conf.checrs_backend_str = server_url
            save_config()

        rest_backend_url = dialog.rest_backend_str
        if rest_backend_url is not None:
            Conf.checrs_rest_endpoint_url = rest_backend_url
            save_config()

    def set_chess_target(self):
        if not self.connected:
            QMessageBox.critical(
                self.workspace.main_window,
                "Backend is not connected",
                "Your angr management instance is not connected to CHECRS backend. Please set a "
                "correct connection string before setting the CHESS target.",
                QMessageBox.Ok,
            )
            return

        if self.workspace.main_instance.project.am_none:
            QMessageBox.critical(
                self.workspace.main_window,
                "No binary is loaded",
                "Please load a binary before associating it to a CHESS target.",
                QMessageBox.Ok,
            )
            return

        dialog = QTargetSelectorDialog(self.workspace, parent=self.workspace.main_window)
        dialog.exec_()

        if dialog.ok:
            if dialog.target_id:
                self.target_id = dialog.target_id
                self.target_image_id = dialog.target_image_id
                self.target_description = dialog.target_description
            else:
                self.target_id = None
                self.target_image_id = None
                self.target_description = None
            self.target_id_updated()

    CONFIG_ENTRIES = [
        ConfigurationEntry("checrs_backend_str", str, "", default_value=""),
        ConfigurationEntry("checrs_rest_endpoint_url", str, "", default_value=""),
    ]

    def angrdb_store_entries(self):
        yield "chess_target_id", self.target_id if self.target_id else ""
        yield "chess_target_description", self.target_description if self.target_description else ""
        yield "chess_target_image_id", self.target_image_id if self.target_image_id else ""

    def angrdb_load_entry(self, key, value):
        if key == "chess_target_id":
            self.target_id = value
            self.target_id_updated()
        elif key == "chess_target_image_id":
            self.target_image_id = value
            self.target_id_updated()
        elif key == "chess_target_description":
            self.target_description = value
            self.target_id_updated()



from typing import TYPE_CHECKING, Optional

import requests
import sqlalchemy
from PySide6.QtGui import Qt
from PySide6.QtWidgets import QDialog, QHBoxLayout, QLabel, QLineEdit, QMessageBox, QPushButton, QVBoxLayout

try:
    import slacrs
except ImportError:
    slacrs = None

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace

    from .chess_connector import ChessConnector


class QBackendSelectorDialog(QDialog):
    """
    Implements a CHESS backend URL input dialog.
    """

    def __init__(
        self,
        workspace: "Workspace",
        backend_str: Optional[str] = None,
        rest_backend_str: Optional[str] = None,
        chess_connector: "ChessConnector" = None,
        parent=None,
    ):
        super().__init__(parent)

        self.workspace = workspace

        self._input: QLineEdit = None
        self._rest_endpoint_input: QLineEdit = None
        self._test_button: QPushButton = None
        self._test_rest_button: QPushButton = None
        self._ok_button: QPushButton = None
        self._cancel_button: QPushButton = None
        self._status_label: QLabel = None

        self.chess_connector = chess_connector
        self.backend_str: Optional[str] = backend_str
        self.rest_backend_str: Optional[str] = rest_backend_str

        self.setWindowTitle("Connect to CHECRS backend")
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        self._init_widgets()

    def _init_widgets(self):
        # input
        input_label = QLabel("CHECRS backend URL:")
        self._input = QLineEdit("sqlite://" if not self.backend_str else self.backend_str)
        self._input.setMinimumWidth(400)
        self._test_button = QPushButton("Test...")
        self._test_button.clicked.connect(self._on_test_button_clicked)
        input_layout = QHBoxLayout()
        input_layout.addWidget(input_label)
        input_layout.addWidget(self._input)
        input_layout.addWidget(self._test_button)

        # REST endpoint URL
        rest_label = QLabel("CHECRS REST backend URL:")
        self._rest_endpoint_input = QLineEdit("http://" if not self.rest_backend_str else self.rest_backend_str)
        self._rest_endpoint_input.setMinimumWidth(400)
        self._test_rest_button = QPushButton("Test...")
        self._test_rest_button.clicked.connect(self._on_test_rest_button_clicked)
        rest_layout = QHBoxLayout()
        rest_layout.addWidget(rest_label)
        rest_layout.addWidget(self._rest_endpoint_input)
        rest_layout.addWidget(self._test_rest_button)

        # status
        self._status_label = QLabel()

        # buttons
        self._ok_button = QPushButton("OK")
        self._ok_button.clicked.connect(self._on_ok_button_clicked)
        self._cancel_button = QPushButton("Cancel")
        self._cancel_button.clicked.connect(self._on_cancel_button_clicked)
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self._ok_button)
        buttons_layout.addWidget(self._cancel_button)

        layout = QVBoxLayout()
        layout.addLayout(input_layout)
        layout.addLayout(rest_layout)
        layout.addWidget(self._status_label)
        layout.addLayout(buttons_layout)

        self.setLayout(layout)

    def test_connection(self, connection_str: str) -> bool:
        if not connection_str:
            return False

        if slacrs is None:
            raise ImportError("The slacrs module is not installed")

        slacrs_instance = self.chess_connector.slacrs_instance(database=connection_str)
        if slacrs_instance is None:
            raise RuntimeError("Cannot create Slacrs instance.")

        # test connection
        session = slacrs_instance.session()
        session.query(sqlalchemy.false()).filter(sqlalchemy.false())
        session.close()

        return True

    @staticmethod
    def test_rest_connection(rest_backend: str) -> bool:
        if not rest_backend:
            return False

        r = requests.head(rest_backend)
        if r.status_code == 404:
            return True

        # unexpected
        return False

    #
    # Events
    #

    def _on_test_button_clicked(self):
        self._test_button.setEnabled(False)

        connection_str = self._input.text().strip(" ")
        self._status_label.setText(f"Testing connection to {connection_str}...")
        self.workspace.main_window.app.processEvents()

        try:
            r = self.test_connection(connection_str)
        except Exception as ex:  # pylint:disable=broad-except
            self._status_label.setText(f"Failed to connect to {connection_str}. Exception: {str(ex)}.")
            r = None

        if r is False:
            self._status_label.setText(f"Failed to connect to {connection_str}.")
        elif r is True:
            self._status_label.setText(f"Successfully connected to {connection_str}.")

        self._test_button.setEnabled(True)

    def _on_test_rest_button_clicked(self):
        self._test_rest_button.setEnabled(False)

        backend = self._rest_endpoint_input.text().strip(" ")
        self._status_label.setText(f"Testing connection to REST backend {backend}...")
        self.workspace.main_window.app.processEvents()

        try:
            r = self.test_rest_connection(backend)
        except Exception as ex:  # pylint:disable=broad-except
            self._status_label.setText(f"Failed to connect to REST backend {backend}. Exception: {str(ex)}.")
            r = None

        if r is False:
            self._status_label.setText(f"Failed to connect to REST backend {backend}.")
        elif r is True:
            self._status_label.setText(f"Successfully connected to REST backend {backend}.")

        self._test_rest_button.setEnabled(True)

    def _on_ok_button_clicked(self):
        # Test the connection
        self._ok_button.setEnabled(False)
        connection_str = self._input.text().strip(" ")

        # special case: if connection_str is empty, it means the user wants to disconnect from CHECRS
        if connection_str:
            self._status_label.setText(f"Testing connection to {connection_str}...")
            self.workspace.main_window.app.processEvents()

            try:
                r = self.test_connection(connection_str)
            except Exception:  # pylint:disable=broad-except
                r = False

            if not r:
                QMessageBox.critical(
                    self,
                    "Connection failed",
                    f"Cannot connect to CHECRS backend {connection_str}. Please check if the backend "
                    f"string is correct.",
                )
                self._status_label.setText("")
                self._ok_button.setEnabled(True)
                return

        self.backend_str = connection_str

        rest_backend_str = self._rest_endpoint_input.text()
        if rest_backend_str and rest_backend_str[-1] != "/":
            rest_backend_str += "/"

        self.rest_backend_str = rest_backend_str
        self.close()

    def _on_cancel_button_clicked(self):
        self.backend_str = None
        self.rest_backend_str = None
        self.close()



import json
import os
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import requests
from PySide6.QtCore import Qt
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QApplication, QFileDialog, QInputDialog, QLineEdit, QMessageBox

from angrmanagement.config import Conf
from angrmanagement.errors import InvalidURLError, UnexpectedStatusCodeError
from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.threads import gui_thread_schedule_async
from angrmanagement.plugins.base_plugin import BasePlugin
from angrmanagement.utils.io import download_url, isurl

from .afl_qemu_bitmap import AFLQemuBitmap
from .chess_trace_list import QChessTraceListDialog
from .multi_trace import MultiTrace
from .qtrace_viewer import QTraceViewer
from .trace_statistics import TraceStatistics

if TYPE_CHECKING:
    from angrmanagement.data.object_container import ObjectContainer


class TraceViewer(BasePlugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.workspace.main_instance.register_container(
            "trace", lambda: None, Optional[TraceStatistics], "The current trace"
        )
        self.workspace.main_instance.register_container(
            "multi_trace", lambda: None, Optional[MultiTrace], "The current set of multiple traces"
        )

        # Register event callbacks
        # self.trace.am_subscribe(self._on_trace_updated)
        self.multi_trace.am_subscribe(self._on_trace_updated)

        self._viewers = []

    def teardown(self):
        # I don't really know a better way to do this. tbh allowing arbitrary widget additions is probably intractable
        for trace_viewer in self._viewers:
            trace_viewer.hide()

    #
    # Forwarding properties
    #

    @property
    def trace(self) -> Union["ObjectContainer", Optional[TraceStatistics]]:
        return self.workspace.main_instance.trace

    @property
    def multi_trace(self) -> Union["ObjectContainer", Optional[MultiTrace], AFLQemuBitmap]:
        return self.workspace.main_instance.multi_trace

    #
    # Event handlers
    #

    def _on_trace_updated(self):
        # redraw disassembly view
        view = self.workspace.view_manager.first_view_in_category("disassembly")
        if view is not None:
            view.redraw_current_graph()
        # refresh function table
        view = self.workspace.view_manager.first_view_in_category("functions")
        if view is not None:
            view.refresh()

    URL_ACTIONS = ["openbitmap", "opentrace"]

    def handle_url_action(self, action, kwargs):
        if action == "openbitmap":
            try:
                base = int(kwargs["base"], 16)
            except (ValueError, KeyError):
                base = None
            gui_thread_schedule_async(GlobalInfo.main_window.bring_to_front)
            gui_thread_schedule_async(
                self.open_bitmap_multi_trace,
                args=(kwargs["path"],),
                kwargs={
                    "base_addr": base,
                },
            )
        elif action == "opentrace":
            try:
                base = int(kwargs["base"], 16)
            except (ValueError, KeyError):
                base = None
            path = kwargs["path"]
            gui_thread_schedule_async(GlobalInfo.main_window.bring_to_front)
            gui_thread_schedule_async(
                self.add_trace,
                kwargs={
                    "trace_path": path,
                    "base_addr": base,
                },
            )
        else:
            raise ValueError("Trace plugin cannot handle url action " + action)

    #
    # features for the disassembly view
    #

    GRAPH_TRACE_LEGEND_WIDTH = 30
    GRAPH_TRACE_LEGEND_SPACING = 20

    def instrument_disassembly_view(self, dview):
        trace_viewer = QTraceViewer(self.workspace, dview, parent=dview)
        self._viewers.append(trace_viewer)

        dview.layout().addWidget(trace_viewer)
        trace_viewer.hide()

    def color_block(self, addr):
        if not self.multi_trace.am_none:
            if isinstance(self.multi_trace.am_obj, MultiTrace):
                if not self.multi_trace.is_active_tab:
                    return None
            return self.multi_trace.get_hit_miss_color(addr)
        return None

    def handle_click_block(self, qblock, event):
        btn = event.button()

        if QApplication.keyboardModifiers() == Qt.ControlModifier and btn == Qt.RightButton:
            if self.multi_trace is not None and self.multi_trace.am_obj is not None:
                the_trace = self.multi_trace.get_any_trace(qblock.addr)
                if the_trace is not None:
                    self.trace.am_obj = TraceStatistics(self.workspace, the_trace, self.multi_trace.base_addr)
                    self.trace.am_event()
                    return True

        return False

    def draw_insn(self, qinsn, painter):
        # legend
        if not self.multi_trace.am_none:
            if isinstance(self.multi_trace.am_obj, MultiTrace):
                if not self.multi_trace.is_active_tab:
                    return  # skip
        strata = self._gen_strata(qinsn.insn.addr)
        if strata is not None:
            legend_x = 0 - self.GRAPH_TRACE_LEGEND_WIDTH - self.GRAPH_TRACE_LEGEND_SPACING
            for i, w in strata:
                color = self.trace.get_mark_color(qinsn.insn.addr, i)
                painter.setPen(color)
                painter.setBrush(color)
                painter.drawRect(legend_x, 0, w, qinsn.height)
                legend_x += w

    def _gen_strata(self, addr):
        if not self.trace.am_none:
            # count is cached in trace.
            count = self.trace.get_count(addr)

            if count > self.GRAPH_TRACE_LEGEND_WIDTH:
                jump = count / self.GRAPH_TRACE_LEGEND_WIDTH
                return [(int(jump * i), 1) for i in range(self.GRAPH_TRACE_LEGEND_WIDTH)]
            elif count > 0:
                width = self.GRAPH_TRACE_LEGEND_WIDTH // count
                remainder = self.GRAPH_TRACE_LEGEND_WIDTH % count
                return [(i, width + 1) for i in range(remainder)] + [(i, width) for i in range(remainder, count)]

        return None

    #
    # features for the functions view
    #

    def color_func(self, func):
        if not self.multi_trace.am_none:
            if isinstance(self.multi_trace.am_obj, MultiTrace):
                if self.multi_trace.is_active_tab:
                    return self.multi_trace.get_percent_color(func)
            elif isinstance(self.multi_trace.am_obj, AFLQemuBitmap):
                return self.multi_trace.get_percent_color(func)
            else:
                # you should not reach here
                raise RuntimeError("Impossible happened")

        if not self.trace.am_none:
            if func.addr in self.trace.func_addr_in_trace:
                return QColor(0xF0, 0xE7, 0xDA)
            return QColor(0xEE, 0xEE, 0xEE)
        return None

    FUNC_COLUMNS = ("Coverage",)

    def extract_func_column(self, func, idx):
        assert idx == 0
        if self.multi_trace.am_none:
            cov = 0
            rend = ""
        else:
            cov = self.multi_trace.get_coverage(func)
            rend = "%.2f%%" % cov

        return cov, rend

    #
    # features for loading traces!
    #

    MENU_BUTTONS = [
        "Open/Add trace...",
        "Clear trace",
        "Open AFL bitmap...",
        "Open inverted AFL bitmap...",
        "Reset AFL bitmap",
        "Open traces from CHECRS...",
    ]
    ADD_TRACE_ID = 0
    RESET_TRACE_ID = 1
    OPEN_AFL_BITMAP_ID = 2
    OPEN_AFL_BITMAP_INVERTED_ID = 3
    RESET_AFL_BITMAP = 4
    OPEN_TRACES_FROM_CHECRS = 5

    def handle_click_menu(self, idx):
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return

        if self.workspace.main_instance.project.am_none:
            return

        mapping = {
            self.ADD_TRACE_ID: self.add_trace,
            self.RESET_TRACE_ID: self.reset_trace,
            self.OPEN_AFL_BITMAP_ID: self.open_bitmap_multi_trace,
            self.OPEN_AFL_BITMAP_INVERTED_ID: self.open_inverted_bitmap_multi_trace,
            self.RESET_AFL_BITMAP: self.reset_bitmap,
            self.OPEN_TRACES_FROM_CHECRS: self.open_traces_from_checrs,
        }

        mapping.get(idx)()

    # def open_trace(self):
    #     trace, base_addr = self._open_json_trace_dialog()
    #     if trace is None or base_addr is None:
    #         return

    #     # self.trace.am_obj = TraceStatistics(self.workspace, trace, base_addr)
    #     # self.trace.am_event()
    #     self.multi_trace.am_obj = MultiTrace(self.workspace)
    #     self.trace.am_obj = self.multi_trace.am_obj.add_trace(trace, base_addr)
    #     self.multi_trace.am_event()
    #     self.trace.am_event()

    def add_trace(self, trace_path=None, base_addr=None):
        trace, base_addr = self._open_trace(trace_path, base_addr)
        if trace is None or base_addr is None:
            return

        self._add_trace(trace, base_addr)

    def reset_trace(self):
        self.trace.am_obj = None
        self.trace.am_event()
        self.multi_trace.am_obj = None
        self.multi_trace.am_event()

    def open_bitmap_multi_trace(self, trace_path=None, base_addr=None):
        r = self._open_bitmap_multi_trace(trace_path, base_addr)
        if r is None:
            return
        trace, base_addr = r
        self.multi_trace.am_obj = AFLQemuBitmap(self.workspace, trace, base_addr)
        self.multi_trace.am_event()

    def open_inverted_bitmap_multi_trace(self, trace_path=None, base_addr=None):
        r = self._open_bitmap_multi_trace(trace_path, base_addr)
        if r is None:
            return
        trace, base_addr = r
        self.multi_trace.am_obj = AFLQemuBitmap(self.workspace, trace, base_addr, bits_inverted=True)
        self.multi_trace.am_event()

    def reset_bitmap(self):
        self.multi_trace.am_obj = None
        self.multi_trace.am_event()

    def _open_bitmap_multi_trace(self, trace_path, base_addr):
        if trace_path is None:
            trace_path = self._open_trace_dialog(tfilter="")
            if trace_path is None:
                return None

        if base_addr is None:
            base_addr = self._open_baseaddr_dialog(0x4000000000)
            if base_addr is None:
                return None

        if isurl(trace_path):
            try:
                trace = download_url(trace_path, parent=self.workspace._main_window, to_file=False)
            except InvalidURLError:
                QMessageBox.critical(
                    self.workspace._main_window,
                    "Downloading failed",
                    "angr management failed to download the file. The URL is invalid.",
                )
                return None
            except UnexpectedStatusCodeError as ex:
                QMessageBox.critical(
                    self.workspace._main_window,
                    "Downloading failed",
                    "angr management failed to retrieve the header of the file. "
                    "The HTTP request returned an unexpected status code %d." % ex.status_code,
                )
                return None
        else:
            with open(trace_path, "rb") as f:
                trace = f.read()

        return trace, base_addr

    def _open_trace(self, trace_path, base_addr):
        if trace_path is None:
            return self._open_json_trace_dialog()
        else:
            if isurl(trace_path):
                try:
                    trace_bytes = download_url(trace_path, parent=self.workspace._main_window, to_file=False)
                    trace = json.loads(trace_bytes)
                except InvalidURLError:
                    QMessageBox.critical(
                        self.workspace._main_window,
                        "Downloading failed",
                        "angr management failed to download the file. The URL is invalid.",
                    )
                    trace = None
                except UnexpectedStatusCodeError as ex:
                    QMessageBox.critical(
                        self.workspace._main_window,
                        "Downloading failed",
                        "angr management failed to retrieve the header of the file. "
                        "The HTTP request returned an unexpected status code %d." % ex.status_code,
                    )
                    trace = None
            else:
                with open(trace_path) as f:
                    trace = json.load(f)

            if base_addr is None:
                base_addr = self._open_baseaddr_dialog(0x4000000000)

            return trace, base_addr

    def _open_trace_dialog(self, tfilter):
        file_path, _ = QFileDialog.getOpenFileName(None, "Open a trace", "", tfilter)
        try:
            with open(file_path, "rb") as f:
                f.read(1)
        except FileNotFoundError:
            return None

        return file_path

    def _open_baseaddr_dialog(self, default_base):
        base_addr, _ = QInputDialog.getText(
            None, "Input Trace Base Address", "Base Address:", QLineEdit.Normal, hex(default_base)
        )
        try:
            return int(base_addr, 16)
        except ValueError:
            return None

    def _open_json_trace_dialog(self) -> Tuple[Optional[List[int]], Optional[int]]:
        # project = self.workspace.instance.project
        trace_file_name = self._open_trace_dialog(tfilter="json (*.json)")

        if trace_file_name is None:
            return None, None

        with open(trace_file_name) as f:
            trace = json.load(f)

        if not isinstance(trace, dict):
            QMessageBox.critical(
                self.workspace._main_window,
                "Incorrect trace format",
                "Failed to open the JSON trace. We expect the JSON trace to be a dict.",
            )
            return None, None
        elif "bb_addrs" not in trace.keys():
            QMessageBox.critical(
                self.workspace._main_window,
                "Incorrect trace format",
                "Failed to open the JSON trace." " We expect the JSON trace to contain the field 'bb_addrs'.",
            )
            return None, None
        elif not isinstance(trace["bb_addrs"], list):
            QMessageBox.critical(
                self.workspace._main_window,
                "Incorrect trace format",
                "Failed to open the JSON trace." "We expect the JSON trace bb_addrs field to be a list of integers.",
            )
            return None, None

        base_addr = self._open_baseaddr_dialog(0x4000000000)

        return trace, base_addr

    def _add_trace(self, trace, base_addr):
        if self.multi_trace.am_obj is None:
            self.multi_trace.am_obj = MultiTrace(self.workspace)
        self.trace.am_obj = self.multi_trace.am_obj.add_trace(trace, base_addr)
        self.multi_trace.am_event()
        self.trace.am_event()

    def open_traces_from_checrs(self):
        if not Conf.checrs_rest_endpoint_url:
            QMessageBox.critical(
                self.workspace.main_window,
                "Unspecified CHECRS REST endpoint",
                "The CHECRS REST endpoint is not specified.",
            )
            return

        connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
        if connector is None:
            return
        target_image_id = connector.target_image_id
        if not target_image_id:
            return

        dialog = QChessTraceListDialog(self.workspace, parent=self.workspace.main_window)
        dialog.exec_()

        traces = []
        failures: List[Tuple[str, str]] = []

        if dialog.trace_ids:
            for input_id, trace_id in dialog.trace_ids:
                # talk to the backend to get the trace
                try:
                    url = os.path.join(
                        Conf.checrs_rest_endpoint_url, f"v1/targets/{target_image_id}/seeds/{input_id}/trace"
                    )
                    r = requests.get(url)
                except Exception as ex:
                    failures.append((trace_id, str(ex)))
                    continue
                data = r.text
                try:
                    trace = json.loads(data)
                except json.JSONDecodeError as ex:
                    failures.append((trace_id, str(ex)))
                    continue
                traces.append(trace)

        for trace in traces:
            self._add_trace(trace, 0x4000000000)

        if failures:
            msg = []
            for trace_id, exc in failures:
                msg.append(f"Trace {trace_id}: {exc}.")
            QMessageBox.warning(
                self.workspace.main_window,
                "Failed to load some traces",
                "angr management failed to load some traces.\n" "Details:\n" + "\n".join(msg),
            )



import asyncio
import threading
from time import sleep
from typing import List

from tornado.platform.asyncio import AnyThreadEventLoopPolicy

try:
    from slacrs import Slacrs
    from slacrs.model import Input, InputTag
except ImportError:
    Slacrs = None


class Seed:
    def __init__(self, seed: Input, id: int):
        self.created_at = seed.created_at
        self.tags: List[str] = [x.value for x in seed.tags]
        self.value: bytes = seed.value
        self._realid = seed.id
        self.id: str = hex(id)[2:].rjust(8, "0")


class SeedTable:
    """
    Multiple POIs
    """

    query_signal = None

    def __init__(self, workspace, query_signal, seed_callback=None):
        self.workspace = workspace
        self.seed_callback = seed_callback
        self.connector = None
        self.slacrs_instance = None
        self.should_exit = False
        self.query_signal = query_signal

        self.init_instance()
        self.has_populated_seeds = False

        self.slacrs_thread = threading.Thread(target=self.listen_for_events)
        self.slacrs_thread.setDaemon(True)
        self.slacrs_thread.start()

    def init_instance(self):
        self.connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")

        if self.connector is None:
            self.workspace.log("Unable to retrieve plugin ChessConnector")
            return False

        self.slacrs_instance = self.connector.slacrs_instance()

        if self.slacrs_instance is None:
            self.workspace.log("Unable to retrieve Slacrs instance")
            return False

        return True

    def listen_for_events(self):
        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
        while not self.connector:
            self.connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
            sleep(1)

        while not self.slacrs_instance:
            self.slacrs_instance = self.connector.slacrs_instance()
            sleep(1)

        while not self.connector.target_image_id:
            sleep(1)

        self.seed_callback(self.get_all_seeds())
        self.has_populated_seeds = True

        prev_target = self.connector.target_image_id
        while not self.should_exit:
            if self.connector.target_image_id != prev_target:
                prev_target = self.connector.target_image_id
                self.seed_callback(self.get_all_seeds())

            new_event_count = self.slacrs_instance.fetch_events()
            for _ in range(new_event_count):
                e = self.slacrs_instance.event_queue.get_nowait()
                session = self.slacrs_instance.session()
                if e.kind == "input":
                    obj = e.get_object(session)
                    if (
                        session.query(Input)
                        .filter_by(id=e.object_id)
                        .filter_by(target_image_id=self.connector.target_image_id)
                        == 1
                    ):
                        seed = session.query(Input).filter_by(obj.object_id).one()
                        self.seed_callback(seed)
                session.close()

    def filter_seeds_by_value(self, value: bytes):
        self.query_signal.querySignal.emit(True)
        session = self.slacrs_instance.session()
        seeds: List[Seed] = []
        if session:
            query = str(session.query(Input.id))
            query += f"\nWHERE POSITION('\\x{value.hex()}'::bytea in input.value) != 0\nORDER BY input.created_at"
            result = session.execute(query)
            seeds = [Seed(session.query(Input).filter_by(id=s_id[0]).first(), idx) for idx, s_id in enumerate(result)]
            session.close()
        return seeds

    def filter_seeds_by_tag(self, tags: List[str] = []) -> List[Seed]:
        self.query_signal.querySignal.emit(True)
        session = self.slacrs_instance.session()
        seeds: List[Seed] = []
        if session:
            result = session.query(Input).join(Input.tags)
            for tag in tags:
                result = result.filter(Input.tags.any(InputTag.value == tag))
            result = result.order_by(Input.created_at).all()
            seeds = [Seed(inp, idx) for idx, inp in enumerate(result)]
        return seeds

    def get_all_seeds(self, filter=None):
        self.query_signal.querySignal.emit(True)
        session = self.slacrs_instance.session()
        seeds: List[Seed] = []
        if session:
            result = (
                session.query(Input)
                .filter_by(target_image_id=self.connector.target_image_id)
                .order_by(Input.created_at)
                .all()
            )
            seeds = [Seed(inp, idx) for idx, inp in enumerate(result)]
            # seeds = list(map(lambda val, i: Seed(val, i), enumerate(result))
            session.close()
        if len(seeds) == 0:
            self.workspace.log("Unable to retrieve seeds for target_image_id: %s" % self.connector.target_image_id)

        return seeds



from PySide6.QtGui import QColor

COLOR_SCHEMES = {
    "Light": {
        "disasm_view_operand_color": QColor(0x00, 0x00, 0x80),
        "disasm_view_operand_constant_color": QColor(0x00, 0x00, 0x80),
        "disasm_view_variable_label_color": QColor(0x00, 0x80, 0x00),
        "disasm_view_operand_highlight_color": QColor(0xFC, 0xEF, 0x00),
        "disasm_view_operand_select_color": QColor(0xFF, 0xFF, 0x00),
        "disasm_view_function_color": QColor(0x00, 0x00, 0xFF),
        "disasm_view_ir_default_color": QColor(0x80, 0x80, 0x80),
        "disasm_view_label_color": QColor(0x00, 0x00, 0xFF),
        "disasm_view_label_highlight_color": QColor(0xF0, 0xF0, 0xBF),
        "disasm_view_target_addr_color": QColor(0x00, 0x00, 0xFF),
        "disasm_view_antitarget_addr_color": QColor(0xFF, 0x00, 0x00),
        "disasm_view_node_shadow_color": QColor(0x00, 0x00, 0x00, 0x00),
        "disasm_view_node_background_color": QColor(0xFA, 0xFA, 0xFA),
        "disasm_view_node_zoomed_out_background_color": QColor(0xDA, 0xDA, 0xDA),
        "disasm_view_node_border_color": QColor(0xF0, 0xF0, 0xF0),
        "disasm_view_node_instruction_selected_background_color": QColor(0xB8, 0xC3, 0xD6),
        "disasm_view_node_address_color": QColor(0x00, 0x00, 0x00),
        "disasm_view_node_mnemonic_color": QColor(0x00, 0x00, 0x80),
        "disasm_view_selected_node_border_color": QColor(0x6B, 0x71, 0x7C),
        "disasm_view_printable_byte_color": QColor(0x00, 0x80, 0x40),
        "disasm_view_printable_character_color": QColor(0x00, 0x80, 0x40),
        "disasm_view_unprintable_byte_color": QColor(0x80, 0x40, 0x00),
        "disasm_view_unprintable_character_color": QColor(0x80, 0x40, 0x00),
        "disasm_view_unknown_byte_color": QColor(0xF0, 0x00, 0x00),
        "disasm_view_unknown_character_color": QColor(0xF0, 0x00, 0x00),
        "disasm_view_back_edge_color": QColor(0xF9, 0xD5, 0x77),
        "disasm_view_true_edge_color": QColor(0x79, 0xCC, 0xCD),
        "disasm_view_false_edge_color": QColor(0xF1, 0x66, 0x64),
        "disasm_view_direct_jump_edge_color": QColor(0x56, 0x5A, 0x5C),
        "disasm_view_exception_edge_color": QColor(0xF9, 0x91, 0x0A),
        "hex_view_selection_color": QColor(0xFF, 0x00, 0x00),
        "hex_view_selection_alt_color": QColor(0xA0, 0xA0, 0xA4),
        "hex_view_data_color": QColor(0x00, 0x00, 0xFF),
        "hex_view_string_color": QColor(0x00, 0xFF, 0xFF),
        "hex_view_instruction_color": QColor(0xFF, 0x00, 0xFF),
        "function_table_color": QColor(0x00, 0x00, 0x00),
        "function_table_syscall_color": QColor(0x00, 0x00, 0x80),
        "function_table_plt_color": QColor(0x00, 0x80, 0x00),
        "function_table_simprocedure_color": QColor(0x80, 0x00, 0x00),
        "function_table_alignment_color": QColor(0x80, 0x00, 0x80),
        "function_table_signature_bg_color": QColor(0xAA, 0xFF, 0xFF),
        "palette_window": QColor(0xEF, 0xEF, 0xEF, 0xFF),
        "palette_windowtext": QColor(0x00, 0x00, 0x00, 0xFF),
        "palette_base": QColor(0xFF, 0xFF, 0xFF, 0xFF),
        "palette_alternatebase": QColor(0xF7, 0xF7, 0xF7, 0xFF),
        "palette_tooltipbase": QColor(0xFF, 0xFF, 0xDC, 0xFF),
        "palette_tooltiptext": QColor(0x00, 0x00, 0x00, 0xFF),
        "palette_text": QColor(0x00, 0x00, 0x00, 0xFF),
        "palette_button": QColor(0xEF, 0xEF, 0xEF, 0xFF),
        "palette_buttontext": QColor(0x00, 0x00, 0x00, 0xFF),
        "palette_brighttext": QColor(0xFF, 0xFF, 0xFF, 0xFF),
        "palette_highlight": QColor(0x30, 0x8C, 0xC6, 0xFF),
        "palette_highlightedtext": QColor(0xFF, 0xFF, 0xFF, 0xFF),
        "palette_disabled_text": QColor(0xBE, 0xBE, 0xBE, 0xFF),
        "palette_disabled_buttontext": QColor(0xBE, 0xBE, 0xBE, 0xFF),
        "palette_disabled_windowtext": QColor(0xBE, 0xBE, 0xBE, 0xFF),
        "palette_light": QColor(0xDF, 0xDF, 0xDF, 0xDF),
        "palette_midlight": QColor(0xCA, 0xCA, 0xCA, 0xFF),
        "palette_dark": QColor(0x9F, 0x9F, 0x9F, 0xFF),
        "palette_mid": QColor(0xB8, 0xB8, 0xB8, 0xFF),
        "palette_shadow": QColor(0x76, 0x76, 0x76, 0xFF),
        "palette_link": QColor(0x00, 0x00, 0xFF, 0xFF),
        "palette_linkvisited": QColor(0xFF, 0x00, 0xFF, 0xFF),
        "feature_map_regular_function_color": QColor(0x00, 0xA0, 0xE8),
        "feature_map_unknown_color": QColor(0x0A, 0x0A, 0x0A),
        "feature_map_delimiter_color": QColor(0x00, 0x00, 0x00),
        "feature_map_data_color": QColor(0xC0, 0xC0, 0xC0),
        "feature_map_string_color": QColor(0x00, 0xF0, 0x80),
        "pseudocode_comment_color": QColor(0x00, 0x80, 0x00, 0xFF),
        "pseudocode_function_color": QColor(0x00, 0x00, 0xFF, 0xFF),
        "pseudocode_quotation_color": QColor(0x00, 0x80, 0x00, 0xFF),
        "pseudocode_keyword_color": QColor(0x00, 0x00, 0x80, 0xFF),
        "pseudocode_highlight_color": QColor(0xFF, 0xFF, 0x00, 0xFF),
        "proximity_node_background_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_node_selected_background_color": QColor(0xCC, 0xCC, 0xCC),
        "proximity_node_border_color": QColor(0xF0, 0xF0, 0xF0),
        "proximity_function_node_text_color": QColor(0xFF, 0x00, 0x00),
        "proximity_string_node_text_color": QColor(0x00, 0x80, 0x00),
        "proximity_integer_node_text_color": QColor(0x00, 0x00, 0x80),
        "proximity_variable_node_text_color": QColor(0x00, 0x00, 0x80),
        "proximity_unknown_node_text_color": QColor(0x00, 0x00, 0x80),
        "proximity_call_node_text_color": QColor(0x00, 0x00, 0xFF),
        "proximity_call_node_text_color_plt": QColor(0x8B, 0x00, 0x8B),
        "proximity_call_node_text_color_simproc": QColor(0x8B, 0x00, 0x8B),
    },
    "Dark": {
        "disasm_view_operand_color": QColor(0xF0, 0xF0, 0x5A),
        "disasm_view_operand_constant_color": QColor(0x34, 0xF0, 0x8C),
        "disasm_view_variable_label_color": QColor(0x34, 0xD4, 0xF0),
        "disasm_view_operand_highlight_color": QColor(0x05, 0x2F, 0x50),
        "disasm_view_operand_select_color": QColor(0x09, 0x50, 0x8D),
        "disasm_view_function_color": QColor(0xC8, 0xC8, 0xC8),
        "disasm_view_ir_default_color": QColor(0x80, 0x80, 0x80),
        "disasm_view_label_color": QColor(0x00, 0xAA, 0xFF),
        "disasm_view_label_highlight_color": QColor(0x2F, 0x2F, 0x25),
        "disasm_view_target_addr_color": QColor(0x00, 0xAA, 0xFF),
        "disasm_view_antitarget_addr_color": QColor(0xFF, 0x00, 0x00),
        "disasm_view_node_shadow_color": QColor(0x00, 0x00, 0x00, 0x4B),
        "disasm_view_node_background_color": QColor(0x3C, 0x3C, 0x3C),
        "disasm_view_node_zoomed_out_background_color": QColor(0x64, 0x64, 0x64),
        "disasm_view_node_border_color": QColor(0x50, 0x50, 0x50),
        "disasm_view_node_instruction_selected_background_color": QColor(0x4C, 0x50, 0x58),
        "disasm_view_node_address_color": QColor(0x2C, 0xC9, 0x76),
        "disasm_view_node_mnemonic_color": QColor(0xE0, 0xE0, 0xE0),
        "disasm_view_selected_node_border_color": QColor(0x6B, 0x71, 0x7C),
        "disasm_view_printable_byte_color": QColor(0x00, 0x80, 0x40),
        "disasm_view_printable_character_color": QColor(0x00, 0x80, 0x40),
        "disasm_view_unprintable_byte_color": QColor(0xBA, 0xBA, 0xBA),
        "disasm_view_unprintable_character_color": QColor(0xBA, 0xBA, 0xBA),
        "disasm_view_unknown_byte_color": QColor(0xF0, 0x00, 0x00),
        "disasm_view_unknown_character_color": QColor(0xF0, 0x00, 0x00),
        "disasm_view_back_edge_color": QColor(0xF9, 0xD5, 0x77),
        "disasm_view_true_edge_color": QColor(0x79, 0xCC, 0xCD),
        "disasm_view_false_edge_color": QColor(0xF1, 0x66, 0x64),
        "disasm_view_direct_jump_edge_color": QColor(0x56, 0x5A, 0x5C),
        "disasm_view_exception_edge_color": QColor(0xF9, 0x91, 0x0A),
        "hex_view_selection_color": QColor(0xFF, 0x00, 0x00),
        "hex_view_selection_alt_color": QColor(0xA0, 0xA0, 0xA4),
        "hex_view_data_color": QColor(0x00, 0x00, 0xFF),
        "hex_view_string_color": QColor(0x00, 0xFF, 0xFF),
        "hex_view_instruction_color": QColor(0xFF, 0x00, 0xFF),
        "function_table_color": QColor(0xE0, 0xE0, 0xE0),
        "function_table_syscall_color": QColor(0x00, 0x00, 0x80),
        "function_table_plt_color": QColor(0x00, 0x80, 0x00),
        "function_table_simprocedure_color": QColor(0x80, 0x00, 0x00),
        "function_table_alignment_color": QColor(0x80, 0x80, 0x00),
        "function_table_signature_bg_color": QColor(0xAA, 0xFF, 0xFF),
        "palette_window": QColor(0x35, 0x35, 0x35),
        "palette_windowtext": QColor(0xFF, 0xFF, 0xFF),
        "palette_base": QColor(0x28, 0x28, 0x28),
        "palette_alternatebase": QColor(0x1D, 0x1D, 0x1D),
        "palette_tooltipbase": QColor(0x35, 0x35, 0x35),
        "palette_tooltiptext": QColor(0xFF, 0xFF, 0xFF),
        "palette_text": QColor(0xE0, 0xE0, 0xE0),
        "palette_button": QColor(0x35, 0x35, 0x35),
        "palette_buttontext": QColor(0xE0, 0xE0, 0xE0),
        "palette_brighttext": QColor(0xFF, 0x00, 0x00),
        "palette_highlight": QColor(0x46, 0x85, 0xE5),
        "palette_highlightedtext": QColor(0xFF, 0xFF, 0xFF),
        "palette_disabled_text": QColor(0x80, 0x80, 0x80),
        "palette_disabled_buttontext": QColor(0x80, 0x80, 0x80),
        "palette_disabled_windowtext": QColor(0x80, 0x80, 0x80),
        "palette_light": QColor(0x46, 0x46, 0x46),
        "palette_midlight": QColor(0x40, 0x40, 0x40),
        "palette_dark": QColor(0x20, 0x20, 0x20),
        "palette_mid": QColor(0x28, 0x28, 0x28),
        "palette_shadow": QColor(0x16, 0x16, 0x16),
        "palette_link": QColor(0x2D, 0xC5, 0x2D).lighter(),
        "palette_linkvisited": QColor(0x2D, 0xC5, 0x2D).darker(),
        "feature_map_regular_function_color": QColor(0x00, 0xA0, 0xE8),
        "feature_map_unknown_color": QColor(0x0A, 0x0A, 0x0A),
        "feature_map_delimiter_color": QColor(0x00, 0x00, 0x00),
        "feature_map_data_color": QColor(0xC0, 0xC0, 0xC0),
        "feature_map_string_color": QColor(0x00, 0xF0, 0x80),
        "pseudocode_comment_color": QColor(0x00, 0x80, 0x00, 0xFF),
        "pseudocode_function_color": QColor(0x60, 0x80, 0xFF, 0xFF),
        "pseudocode_quotation_color": QColor(0x00, 0x80, 0x00, 0xFF),
        "pseudocode_keyword_color": QColor(0x00, 0xFF, 0xFF, 0xFF),
        "pseudocode_highlight_color": QColor(0x59, 0x44, 0x05, 0xFF),
        "proximity_node_background_color": QColor(0x3C, 0x3C, 0x3C),
        "proximity_node_selected_background_color": QColor(0x4C, 0x50, 0x58),
        "proximity_node_border_color": QColor(0x50, 0x50, 0x50),
        "proximity_function_node_text_color": QColor(0xFF, 0x00, 0xFF),
        "proximity_string_node_text_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_integer_node_text_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_variable_node_text_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_unknown_node_text_color": QColor(0xFA, 0xFA, 0xFA),
        "proximity_call_node_text_color": QColor(0x34, 0xD4, 0xF0),
        "proximity_call_node_text_color_plt": QColor(0x2C, 0xC9, 0x76),
        "proximity_call_node_text_color_simproc": QColor(0x2C, 0xC9, 0x76),
    },
}



import logging
from copy import deepcopy
from uuid import uuid4

from PySide6.QtCore import QPoint, Qt
from PySide6.QtGui import QBrush, QColor, QContextMenuEvent, QFont, QImage, QLinearGradient, QPainter, QPen
from PySide6.QtWidgets import (
    QAbstractItemView,
    QGraphicsItemGroup,
    QGraphicsScene,
    QGraphicsView,
    QHeaderView,
    QMenu,
    QMessageBox,
    QTableWidget,
    QTableWidgetItem,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from angrmanagement.ui.views.view import BaseView

from .multi_poi import MultiPOI
from .trace_statistics import TraceStatistics

_l = logging.getLogger(name=__name__)

EMPTY_POI = {
    "category": "",
    "output": {"function": None, "bbl": None, "base": None, "bbl_history": None, "diagnose": None},
}


class QPOIViewer(QWidget):
    """
    POI Viewer QWidget
    """

    TAG_SPACING = 50
    LEGEND_X = -50
    LEGEND_Y = 0
    LEGEND_WIDTH = 10

    TRACE_FUNC_X = 0
    TRACE_FUNC_Y = 0
    TRACE_FUNC_WIDTH = 50
    TRACE_FUNC_MINHEIGHT = 1000

    TAB_HEADER_SIZE = 40
    MAX_WINDOW_SIZE = 500

    MARK_X = LEGEND_X
    MARK_WIDTH = TRACE_FUNC_X - LEGEND_X + TRACE_FUNC_WIDTH
    MARK_HEIGHT = 1

    POIID_COLUMN = 0
    CRASH_COLUMN = 1
    CATEGORY_COLUMN = 2
    DIAGNOSE_COLUMN = 3
    COLUMN_FIELD = ["id", "bbl", "category", "diagnose"]

    def __init__(self, workspace, parent=None, diagnose_handler=None):
        super().__init__(parent=parent)
        self.workspace = workspace
        self._diagnose_handler = diagnose_handler

        self.mark = None
        self.legend = None
        self.legend_height = 0
        self.legend_img = None
        self.trace_func_unit_height = 0

        self.trace_func = None
        self.trace_id = None

        self.tabView = None
        self.traceView = None
        self.traceScene = None
        self.POITraceTab = None
        self.multiPOITab: QWidget = None
        self.multiPOIList: QTableWidget = None

        self.mark = None
        self.curr_position = 0
        self._use_precise_position = False
        self._selected_traces = []
        self._selected_poi = None

        self._init_widgets()

        self.selected_ins.am_subscribe(self._subscribe_select_ins)
        self.poi_trace.am_subscribe(self._subscribe_set_trace)
        self.multi_poi.am_subscribe(self._subscribe_add_poi)

        self.multiPOIList.cellDoubleClicked.connect(self._on_cell_double_click)
        self.multiPOIList.itemChanged.connect(self._on_diagnose_change)

    #
    # Forwarding properties
    #

    @property
    def disasm_view(self):
        """
        Get the current disassembly view (if there is one), or create a new one as needed.
        """
        view = self.workspace.view_manager.current_view_in_category("disassembly")
        if view is None:
            view = self.workspace._get_or_create_disassembly_view()
        return view

    @property
    def poi_trace(self):
        return self.workspace.main_instance.poi_trace

    @property
    def multi_poi(self):
        return self.workspace.main_instance.multi_poi

    @property
    def selected_ins(self):
        return self.disasm_view.infodock.selected_insns

    def _init_widgets(self):
        _l.debug("QPOI Viewer Initiating")
        self.tabView = QTabWidget()  # QGraphicsView()
        self.tabView.setContentsMargins(0, 0, 0, 0)

        #
        # POI trace Tab
        #
        self.POITraceTab = QWidget()
        self.POITraceTab.setContentsMargins(0, 0, 0, 0)
        singleLayout = QVBoxLayout()
        singleLayout.setSpacing(0)
        singleLayout.setContentsMargins(0, 0, 0, 0)

        self.traceView = QGraphicsView()
        self.traceScene = QGraphicsScene()
        self.traceView.setScene(self.traceScene)

        singleLayout.addWidget(self.traceView)
        self.POITraceTab.setLayout(singleLayout)

        #
        # multiPOI Tab
        #
        self.multiPOITab = QMultiPOITab(self)
        # self.multiPOITab = QWidget()
        multiLayout = QVBoxLayout()
        multiLayout.setSpacing(0)
        multiLayout.setContentsMargins(0, 0, 0, 0)

        self.multiPOIList = QTableWidget(0, 4)  # row, col
        self.multiPOIList.setHorizontalHeaderItem(0, QTableWidgetItem("ID"))
        self.multiPOIList.setHorizontalHeaderItem(1, QTableWidgetItem("Crash Point"))
        self.multiPOIList.setHorizontalHeaderItem(2, QTableWidgetItem("Tag"))
        self.multiPOIList.setHorizontalHeaderItem(3, QTableWidgetItem("Diagnose"))
        self.multiPOIList.horizontalHeader().setStretchLastSection(True)
        self.multiPOIList.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.multiPOIList.setSelectionBehavior(QAbstractItemView.SelectRows)
        multiLayout.addWidget(self.multiPOIList)
        self.multiPOITab.setLayout(multiLayout)

        self.tabView.addTab(self.multiPOITab, "POI List")
        self.tabView.addTab(self.POITraceTab, "POI Trace")

        self.POI_TRACE = 1
        self.MULTI_POI = 0

        layout = QVBoxLayout()
        layout.addWidget(self.tabView)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)
        self.show()

    def _reset(self):
        self.traceScene.clear()  # clear items
        self.mark = None

        self.legend = None
        self.legend_height = 0

        self.trace_func = QGraphicsItemGroup()
        self.trace_id = QGraphicsItemGroup()
        self.traceScene.addItem(self.trace_func)
        self.hide()

    #
    # Event
    #

    def _on_cell_double_click(self, row, _):
        _l.debug("row %d is double clicked", row)
        first_cell = self.multiPOIList.item(row, 0)
        if first_cell is None:
            return
        poi_id = first_cell.text()
        poi = self.multi_poi.am_obj.get_poi_by_id(poi_id)
        if poi is None:
            return
        # sanity checks
        if not isinstance(poi, dict):
            return
        if "output" not in poi or not isinstance(poi["output"], dict):
            return
        if "bbl_history" not in poi["output"]:
            return

        trace = poi["output"]["bbl_history"]
        if self._selected_poi != poi_id and trace is not None:
            # render the trace
            self.poi_trace.am_obj = TraceStatistics(self.workspace, trace, trace_id=poi_id)

            # show the trace statistic in POI trace
            self.poi_trace.am_event()

            # show covered basic blocks and functions
            self.multi_poi.am_obj.reload_heatmap(poi_id)

            # redraw function view
            view = self.workspace.view_manager.first_view_in_category("functions")
            if view is not None:
                view.refresh()

            # redraw disassembly view
            view = self.workspace.view_manager.first_view_in_category("disassembly")
            if view is not None:
                view.redraw_current_graph()

        if trace is not None:
            # switch to POI trace tab
            self.tabView.setCurrentIndex(self.POI_TRACE)
        self._selected_poi = poi_id

        second_cell = self.multiPOIList.item(row, 1)
        crash_addr = None
        if second_cell is not None:
            try:
                crash_addr = int(second_cell.text(), 16)
            except ValueError:
                pass
        if crash_addr is not None:
            # show the crashing address
            view = self.workspace.view_manager.first_view_in_category("disassembly")
            if view is not None:
                crash_func = self._get_func_from_addr(crash_addr)
                if crash_func is not None:
                    self.workspace.on_function_selected(crash_func)
                    self.selected_ins.clear()
                    self.selected_ins.update([crash_addr])
                    self.selected_ins.am_event()
                    view.current_graph.show_instruction(crash_addr)

    def _on_diagnose_change(self, item: QTableWidgetItem):
        column = item.column()
        row = item.row()

        poi_id = self.multiPOIList.item(row, self.POIID_COLUMN).text()
        content = item.text()
        original_content = self.multi_poi.am_obj.get_content_by_id_column(poi_id, column)
        _l.debug("updaing %s, content: %s, original: %s", poi_id, content, original_content)
        if not self._is_identical(content, original_content):
            updated_poi = self.multi_poi.update_poi(poi_id, column, content)
            self._diagnose_handler.submit_updated_poi(poi_id, updated_poi)

    def _subscribe_add_poi(self):
        _l.debug("add a poi to multi poi list")
        if self.multi_poi.am_none:
            self.multi_poi.am_obj = MultiPOI(self.workspace)
        poi_ids = self.multi_poi.am_obj.get_all_poi_ids()

        self.multiPOIList.clearContents()
        self._populate_poi_table(self.multiPOIList, poi_ids)
        self.show()

    def _subscribe_set_trace(self):
        _l.debug("on set trace in poi trace viewer")
        self._reset()
        if self.poi_trace.am_none:
            return
        _l.debug("minheight: %d, count: %d", self.TRACE_FUNC_MINHEIGHT, self.poi_trace.count)
        if self.poi_trace.count <= 0:
            _l.warning("No valid addresses found in trace to show. Check base address offsets?")
            self.poi_trace.am_obj = None
            self.poi_trace.am_event()
            return
        if self.TRACE_FUNC_MINHEIGHT < self.poi_trace.count * 15:
            self.trace_func_unit_height = 15
            show_func_tag = True
        else:
            self.trace_func_unit_height = self.TRACE_FUNC_MINHEIGHT / self.poi_trace.count
            show_func_tag = True

        self.legend_height = int(self.poi_trace.count * self.trace_func_unit_height)

        self._show_trace_func(show_func_tag)
        self._show_legend()
        self._set_mark_color()
        self._refresh_multi_list()

        # boundingSize = self.traceScene.itemsBoundingRect().width()
        # windowSize = boundingSize
        # if boundingSize > self.MAX_WINDOW_SIZE:
        #     windowSize = self.MAX_WINDOW_SIZE
        # self.traceScene.setSceneRect(self.traceScene.itemsBoundingRect()) #resize
        # if windowSize > self.width():
        #     self.setMinimumWidth(windowSize)

        self.show()

    def _subscribe_select_ins(self, **kwargs):  # pylint: disable=unused-argument
        if self.poi_trace.am_none:
            return

        if self.mark is not None:
            for i in self.mark.childItems():
                self.mark.removeFromGroup(i)
                self.traceScene.removeItem(i)
            self.traceScene.removeItem(self.mark)

        self.mark = QGraphicsItemGroup()
        self.traceScene.addItem(self.mark)

        if self.selected_ins:
            addr = next(iter(self.selected_ins))
            positions = self.poi_trace.get_positions(addr)
            if positions:  # if addr is in list of positions
                # handle case where insn was selected from disas view
                if not self._use_precise_position:
                    self.curr_position = positions[0] - self.poi_trace.count
                for p in positions:
                    color = self._get_mark_color(p, self.poi_trace.count)
                    y = self._get_mark_y(p)

                    if p == self.poi_trace.count + self.curr_position:  # add thicker line for 'current' mark
                        self.mark.addToGroup(
                            self.traceScene.addRect(
                                self.MARK_X,
                                y,
                                self.MARK_WIDTH,
                                self.MARK_HEIGHT * 4,
                                QPen(QColor("black")),
                                QBrush(color),
                            )
                        )
                    else:
                        self.mark.addToGroup(
                            self.traceScene.addRect(
                                self.MARK_X, y, self.MARK_WIDTH, self.MARK_HEIGHT, QPen(color), QBrush(color)
                            )
                        )

                self.traceScene.update()  # force redraw of the traceScene
                self.scroll_to_position(self.curr_position)

    def _get_func_from_addr(self, addr):
        if self.workspace.main_instance.cfg.am_none:
            return None
        bbl = self.workspace.main_instance.cfg.get_any_node(addr, anyaddr=True)
        function_addr = bbl.function_address
        return self.workspace.main_instance.project.kb.functions.get(function_addr)

    def _populate_poi_table(self, view, poi_ids):
        view.clearContents()
        view.setRowCount(len(poi_ids))
        row = 0  # start after label row
        for poi_id in poi_ids:
            poi = self.multi_poi.am_obj.get_poi_by_id(poi_id)
            _l.debug("populating poi: %s", poi)
            category = poi["category"]
            output = poi["output"]
            crash_addr = output["bbl"]
            if crash_addr is not None:
                crash = hex(crash_addr)
            else:
                crash = None
            diagnose = output.get("diagnose")
            _l.debug("poi_ids: %s", poi_ids)
            _l.debug("current poi id: %s", poi_id)
            self._set_item(view, row, self.POIID_COLUMN, poi_id, editable=False)
            self._set_item(view, row, self.CRASH_COLUMN, crash, editable=True)
            self._set_item(view, row, self.CATEGORY_COLUMN, category, editable=True)
            self._set_item(view, row, self.DIAGNOSE_COLUMN, diagnose, editable=True)
            row += 1
            _l.debug("poi_ids: %s", poi_ids)

    @staticmethod
    def _set_item(view, row, column, text, editable=True):
        if not text:
            text = ""
        item = QTableWidgetItem(text)
        if not editable:
            item.setFlags(item.flags() ^ Qt.ItemIsEditable)
        view.setItem(row, column, item)

    def _refresh_multi_list(self):
        multiPOI = self.multi_poi.am_obj
        trace_ids = multiPOI.get_all_poi_ids()

        self.multiPOIList.clearContents()
        self._populate_poi_table(self.multiPOIList, trace_ids)
        if self._selected_traces and self.multiPOIList.rowCount() > 0:
            self.multiPOIList.item(0, 0).setSelected(True)
            self.multiPOIList.item(0, 1).setSelected(True)
        else:
            for row in range(self.multiPOIList.rowCount()):
                item = self.multiPOIList.item(row, 0)
                inputItem = self.multiPOIList.item(row, 1)
                if item.text() in self._selected_traces:
                    item.setSelected(True)
                    inputItem.setSelected(True)
        self.multi_poi.am_event()

    def _on_tab_change(self):
        multiPOI = self.multi_poi.am_obj
        if self.tabView.currentIndex() == self.MULTI_POI:
            multiPOI.is_active_tab = True
            self._refresh_multi_list()
        elif self.tabView.currentIndex() == self.POI_TRACE:
            multiPOI = self.multi_poi.am_obj
            multiPOI.is_active_tab = False
            # self._show_trace_ids()

    def scroll_to_position(self, position):
        relative_pos = self.poi_trace.count + position
        y_offset = self._get_mark_y(relative_pos)

        scrollValue = 0
        if y_offset > 0.5 * self.traceView.size().height():
            scrollValue = y_offset - 0.5 * self.traceView.size().height()
        scrollValue = min(scrollValue, self.traceView.verticalScrollBar().maximum())
        self.traceView.verticalScrollBar().setValue(scrollValue)
        self._use_precise_position = False

    def jump_next_insn(self):
        # for some reason indexing is done backwards
        if self.curr_position + self.poi_trace.count < self.poi_trace.count - 1:
            self.curr_position += 1
            self._use_precise_position = True
            bbl_addr = self.poi_trace.get_bbl_from_position(self.curr_position)
            func = self.poi_trace.get_func_from_position(self.curr_position)
            self._jump_bbl(func, bbl_addr)

    def jump_prev_insn(self):
        if self.curr_position + self.poi_trace.count > 0:
            self.curr_position -= 1
            self._use_precise_position = True
            bbl_addr = self.poi_trace.get_bbl_from_position(self.curr_position)
            func = self.poi_trace.get_func_from_position(self.curr_position)
            self._jump_bbl(func, bbl_addr)

    def mousePressEvent(self, event):
        button = event.button()
        pos = self._to_logical_pos(event.pos())
        if button == Qt.LeftButton and self.tabView.currentIndex() == self.POI_TRACE and self._at_legend(pos):
            func = self._get_func_from_y(pos.y())
            bbl_addr = self._get_bbl_from_y(pos.y())
            self._use_precise_position = True
            self.curr_position = self._get_position(pos.y())
            self._jump_bbl(func, bbl_addr)

    def _jump_bbl(self, func, bbl_addr):
        disasm_view = self.disasm_view
        if disasm_view is not None:
            all_insn_addrs = self.workspace.main_instance.project.factory.block(bbl_addr).instruction_addrs
            # TODO: replace this with am_events perhaps?
            self.workspace.on_function_selected(func)
            self.selected_ins.clear()
            self.selected_ins.update(all_insn_addrs)
            self.selected_ins.am_event()
            # TODO: this ought to happen automatically as a result of the am_event
            disasm_view.current_graph.show_instruction(bbl_addr)

    def _get_mark_color(self, i, total):
        relative_gradient_pos = i * 1000 // total
        return self.legend_img.pixelColor(self.LEGEND_WIDTH // 2, relative_gradient_pos)

    def _get_mark_y(self, i):
        return self.TRACE_FUNC_Y + self.trace_func_unit_height * i

    def _show_trace_func(self, show_func_tag=True):
        x = self.TRACE_FUNC_X
        y = self.TRACE_FUNC_Y
        prev_name = None
        for position in self.poi_trace.trace_func:
            func_name = position.func_name
            color = self.poi_trace.get_func_color(func_name)
            self.trace_func.addToGroup(
                self.traceScene.addRect(
                    x, y, self.TRACE_FUNC_WIDTH, self.trace_func_unit_height, QPen(color), QBrush(color)
                )
            )
            if show_func_tag is True and func_name != prev_name:
                tag = self.traceScene.addText(func_name, QFont("Source Code Pro", 7))
                tag.setPos(x + self.TRACE_FUNC_WIDTH + self.TAG_SPACING, y - tag.boundingRect().height() // 2)
                self.trace_func.addToGroup(tag)
                anchor = self.traceScene.addLine(
                    self.TRACE_FUNC_X + self.TRACE_FUNC_WIDTH, y, x + self.TRACE_FUNC_WIDTH + self.TAG_SPACING, y
                )
                self.trace_func.addToGroup(anchor)
                prev_name = func_name
            y += self.trace_func_unit_height

    @staticmethod
    def _make_legend_gradient(x1, y1, x2, y2):
        gradient = QLinearGradient(x1, y1, x2, y2)
        gradient.setColorAt(0.0, Qt.red)
        gradient.setColorAt(0.4, Qt.yellow)
        gradient.setColorAt(0.6, Qt.green)
        gradient.setColorAt(0.8, Qt.blue)
        gradient.setColorAt(1.0, Qt.darkBlue)
        return gradient

    def _show_legend(self):
        pen = QPen(Qt.transparent)

        gradient = self._make_legend_gradient(
            self.LEGEND_X, self.LEGEND_Y, self.LEGEND_X, self.LEGEND_Y + self.legend_height
        )
        brush = QBrush(gradient)
        self.legend = self.traceScene.addRect(
            self.LEGEND_X, self.LEGEND_Y, self.LEGEND_WIDTH, self.legend_height, pen, brush
        )

        reference_gradient = self._make_legend_gradient(0, 0, self.LEGEND_WIDTH, 1000)
        base_img = QImage(self.LEGEND_WIDTH, 1000, QImage.Format.Format_ARGB32)
        p = QPainter(base_img)
        p.fillRect(base_img.rect(), reference_gradient)
        self.legend_img = base_img  # reference shade

    def _set_mark_color(self):
        _l.debug("trace count: %d", self.poi_trace.count)
        for p in range(self.poi_trace.count):
            color = self._get_mark_color(p, self.poi_trace.count)
            self.poi_trace.set_mark_color(p, color)

    def _at_legend(self, pos):
        x = pos.x()
        y = pos.y()
        return (
            self.TRACE_FUNC_X + self.LEGEND_X < x < self.traceView.width()
            and self.TRACE_FUNC_Y < y < self.TRACE_FUNC_Y + self.legend_height
        )

    def _to_logical_pos(self, pos):
        x_offset = self.traceView.horizontalScrollBar().value()
        y_offset = self.traceView.verticalScrollBar().value()
        return QPoint(pos.x() + x_offset, pos.y() + y_offset)

    def _get_position(self, y):
        y_relative = y - self.legend_height - self.TAB_HEADER_SIZE

        return int(y_relative // self.trace_func_unit_height)

    def _get_bbl_from_y(self, y):
        position = self._get_position(y)
        return self.poi_trace.get_bbl_from_position(position)

    def _get_func_from_y(self, y):
        position = self._get_position(y)
        func = self.poi_trace.get_func_from_position(position)
        return func

    #
    # Context Menu
    #

    def menu_add_empty_poi(self):
        _l.debug("adding a new empty poi item")

        if self._diagnose_handler.get_image_id() is None:
            QMessageBox.warning(
                self.workspace.main_window,
                "No CHESS target available",
                "No angr project is loaded, or you did not associate the current project with a CHESS "
                "target. Please load a binary and associate it with a CHESS target before creating "
                "POIs.",
            )
            return

        poi_id = str(uuid4())
        if self.multi_poi.am_none:
            self.multi_poi.am_obj = MultiPOI(self.workspace)
        empty_poi = deepcopy(EMPTY_POI)
        self.multi_poi.add_poi(poi_id, empty_poi)
        self.multi_poi.am_event()
        self._diagnose_handler.submit_updated_poi(poi_id, empty_poi)

    def menu_remove_poi(self):
        items = self.multiPOIList.selectedItems()
        row = items.pop().row()
        poi_id = self.multiPOIList.item(row, 0).text()
        _l.debug("removing ID %s", poi_id)
        self.multi_poi.remove_poi(poi_id)
        self.multi_poi.am_event()

    @staticmethod
    def _is_identical(content, original_content):
        if content == original_content:
            return True
        if content == "" and original_content is None:
            return True
        try:
            if int(content, 16) == int(original_content):
                return True
        except (TypeError, ValueError):
            return False
        return False


class QMultiPOITab(QWidget):
    """
    The tab widget for multi POIs
    """

    def __init__(self, poi_view: QPOIViewer):
        super().__init__()
        self.POIView = poi_view

    def contextMenuEvent(self, event: QContextMenuEvent):
        menu = QMenu("", self)
        menu.addAction("Add POI", self.POIView.menu_add_empty_poi)

        remove = menu.addAction("Remove", self.POIView.menu_remove_poi)
        if len(self.POIView.multiPOIList.selectedItems()) == 0:
            remove.setDisabled(True)

        # remove = menu.addAction('Remove both locally and remotely...', self.POIView.menu_remove_poi_remotely)
        # if len(self.POIView.multiPOIList.selectedItems()) == 0:
        #     remove.setDisabled(True)

        menu.exec_(event.globalPos())


class POIView(BaseView):
    """
    The view for displaying POIs.
    """

    def __init__(self, workspace, default_docking_position, diagnose_handler, *args, **kwargs):
        super().__init__("poi", workspace, default_docking_position, *args, **kwargs)

        self.base_caption = "Points of interest"
        self._diagnose_handler = diagnose_handler

        self._poi_viewer: QPOIViewer = None

        self._init_widgets()

        self.width_hint = 250

    def _init_widgets(self):
        self._poi_viewer = QPOIViewer(self.workspace, diagnose_handler=self._diagnose_handler)
        layout = QVBoxLayout()
        layout.addWidget(self._poi_viewer)
        self.setLayout(layout)



# pylint:disable=missing-class-docstring
import datetime
import json
import threading
from time import sleep
from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QAbstractItemView,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QTableWidget,
    QTableWidgetItem,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)
from sqlalchemy import func as sqlalchemy_func

from angrmanagement.logic.threads import gui_thread_schedule_async
from angrmanagement.ui.views.view import BaseView

try:
    from slacrs import Slacrs
    from slacrs.model import PluginMessage
except ImportError:
    Slacrs = None

if TYPE_CHECKING:
    from .chess_connector import ChessConnector


#
# Plugin Log
#


class QPluginItem:
    def __init__(self, plugin, kind, image_id, msg):
        self.plugin = plugin
        self.kind = kind
        self.image_id = image_id
        self.msg = msg

    def widgets(self):
        widgets = [
            QTableWidgetItem(self.plugin),
            QTableWidgetItem(self.kind),
            QTableWidgetItem(self.image_id),
            QTableWidgetItem(self.msg),
        ]

        for w in widgets:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)

        return widgets


class QPluginsLogTable(QTableWidget):
    HEADER = [
        "Plugin",
        "Kind",
        "Image ID",
        "Message",
    ]

    def __init__(self, workspace, parent=None):
        super().__init__(parent)
        self.workspace = workspace

        self.setColumnCount(len(self.HEADER))
        self.setHorizontalHeaderLabels(self.HEADER)
        self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(24)

        self.items = []
        self.cnt = 0

    def reload(self):
        self.setRowCount(len(self.items))

        for idx, item in enumerate(self.items):
            for i, it in enumerate(item.widgets()):
                self.setItem(idx, i, it)

        self.viewport().update()

    def update_table(self, messages):
        self.items = []

        for msg in messages:
            self.items.append(QPluginItem(msg.plugin, msg.kind, msg.target_image_id, msg.message))

        self.reload()


#
#
#


class QFuzztainerItem:
    def __init__(self, data):
        self.data = data

    def widgets(self):
        widgets = [
            QTableWidgetItem(self.data),
        ]

        for w in widgets:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)

        return widgets


class QFuzztainerTable(QTableWidget):
    DISPLAY_ORDER = [
        "execs_per_sec",
        "unique_crashes",
        "unique_hangs",
        "paths_found",
        "paths_total",
        "testcache_size",
        "testcache_count",
        "testcache_evict",
        "run_time",
        "cycles_done",
        "cycles_wo_finds",
        "execs_done",
        "paths_favored",
        "max_depth",
        "pending_favs",
        "pending_total",
        "variable_paths",
        "execs_since_crash",
        "slowest_exec_ms",
        "edges_found",
    ]

    HEADER = [
        "execs/sec",
        "unique crashes",
        "unique hangs",
        "paths found",
        "paths total",
        "testcache size",
        "testcache count",
        "testcache evict",
        "run time",
        "cycles done",
        "cycles w/o finds",
        "execs done",
        "paths favored",
        "max depth",
        "pending favs",
        "pending total",
        "variable paths",
        "execs since crash",
        "slowest exec ms",
        "edges found",
    ]

    def __init__(self, workspace, idx, parent=None):
        super().__init__(parent)
        self.workspace = workspace
        self.idx = idx

        self.setRowCount(len(self.HEADER[self.idx * 10 : self.idx * 10 + 10]))
        self.setVerticalHeaderLabels(self.HEADER[self.idx * 10 : self.idx * 10 + 10])
        self.setColumnCount(1)

        self.verticalHeader().setVisible(True)
        self.horizontalHeader().setVisible(False)

        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)

        self.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.verticalHeader().setSectionResizeMode(QHeaderView.Stretch)
        # self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)

        self.items = []

    def reload(self):
        self.setRowCount(len(self.items))

        for idx, item in enumerate(self.items):
            for i, it in enumerate(item.widgets()):
                self.setItem(idx, i, it)

        self.viewport().update()

    def update_table(self, messages):
        if messages is None:
            return

        self.items = []

        messages = messages[::-1]
        for msg in messages:
            if msg.plugin.lower() == "fuzztainer" and msg.kind == "info":
                try:
                    stats = json.loads(msg.message)
                except Exception as ex:  # pylint:disable=broad-except
                    print(ex)
                    continue

                try:
                    stats["execs_per_sec"]
                except KeyError as e:
                    print(e)
                    continue

                for stat_name in self.DISPLAY_ORDER[self.idx * 10 : self.idx * 10 + 10]:
                    stat_data = stats[stat_name]
                    self.items.append(QFuzztainerItem(stat_data))

                break

        self.reload()


class SummaryView(BaseView):
    def __init__(self, workspace, default_docking_position, connector: "ChessConnector", *args, **kwargs):
        super().__init__("chess_summary", workspace, default_docking_position, *args, **kwargs)
        self.base_caption = "CHESS"

        self.workspace = workspace
        self.connector = connector

        self._init_widgets()

        # start the worker routine
        self.should_work = Slacrs is not None
        self.worker_thread = threading.Thread(target=self.worker_routine)
        self.worker_thread.setDaemon(True)
        self.worker_thread.start()

    def teardown(self):
        self.should_work = False
        self.close()

    #
    # UI
    #

    def _init_widgets(self):
        self.fuzztainer_table_1 = QFuzztainerTable(self.workspace, 0)
        self.fuzztainer_table_2 = QFuzztainerTable(self.workspace, 1)
        self.log_table = QPluginsLogTable(self.workspace)

        #
        # fuzztainer tab
        #

        self.fuzztainer_tab = QWidget()
        fuzz_layout = QHBoxLayout()
        fuzz_layout.addWidget(self.fuzztainer_table_1)
        fuzz_layout.addWidget(self.fuzztainer_table_2)
        self.fuzztainer_tab.setLayout(fuzz_layout)

        #
        # plugin log tab
        #

        self.log_tab = QWidget()
        log_layout = QHBoxLayout()
        log_layout.addWidget(self.log_table)
        self.log_tab.setLayout(log_layout)

        #
        # final tab setup
        #

        # tabs for multiple summaries
        self.tabView = QTabWidget()
        self.tabView.addTab(self.fuzztainer_tab, "Fuzztainer")
        self.tabView.addTab(self.log_tab, "Plugins Log")

        layout = QVBoxLayout()
        layout.addWidget(self.tabView)

        #
        # status bar
        #
        self.status_bar = QLabel()

        layout.addWidget(self.status_bar)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.show()

    def set_status(self, text: str):
        self.status_bar.setText(text)

    #
    # Data Threading
    #

    def worker_routine(self):
        while self.should_work:
            if self.connector is None or not self.connector.connected:
                gui_thread_schedule_async(self.set_status, args=("Not connected to CHECRS backend.",))
                sleep(10)
                continue

            if not self.connector.target_image_id:
                gui_thread_schedule_async(self.set_status, args=("No associated remote target.",))
                sleep(2)
                continue

            sleep(5)

            gui_thread_schedule_async(self.set_status, args=("Refreshing...",))
            session = None
            try:
                session = self._init_session()

                if not session:
                    continue

                self._update_tables(session, self.connector.target_image_id)
            finally:
                if session is not None:
                    session.close()

            last_updated = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            gui_thread_schedule_async(self.set_status, args=(f"Ready. Last updated at {last_updated}.",))

    def _init_session(self):
        if self.connector is None:
            self.connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
            if self.connector is None:
                return None

        try:
            slacrs_instance = self.connector.slacrs_instance()
        except Exception:
            return None

        if slacrs_instance is None:
            return None

        session = slacrs_instance.session()
        return session

    @staticmethod
    def _find_latest_fuzztainer_stats(messages):
        messages = messages[::-1]
        for msg in messages:
            if msg.plugin.lower() == "fuzztainer" and msg.kind == "info":
                try:
                    stats = json.loads(msg.message)
                except Exception as e:
                    print(e)
                    continue

                try:
                    stats["execs_per_sec"]
                except KeyError as e:
                    print(e)
                    continue

                return [msg]
        return None

    def _update_tables(self, session, target_image_id: str):
        res = (
            session.query(PluginMessage)
            .filter(
                sqlalchemy_func.lower(PluginMessage.plugin) == "fuzztainer",
                PluginMessage.kind == "info",
                PluginMessage.target_image_id == target_image_id,
            )
            .order_by(PluginMessage.created_at.desc())
            .limit(50)
        )
        if not res:
            return

        messages = res.all()
        if not messages:
            return

        fuzztainer_msgs = self._find_latest_fuzztainer_stats(messages)
        self.log_table.update_table(messages)
        self.fuzztainer_table_1.update_table(fuzztainer_msgs)
        self.fuzztainer_table_2.update_table(fuzztainer_msgs)
